import io
from pathlib import Path

import pytest

from dissect.target.filesystem import VirtualFilesystem
from dissect.target.loaders.multiraw import MultiRawLoader


def test_multiraw_loader(tmp_path: Path, monkeypatch: pytest.MonkeyPatch):
    (tmp_path / "file1.bin").touch()
    (tmp_path / "file2.bin").touch()

    root = str(tmp_path)
    assert not MultiRawLoader.detect(Path(f"{root}/file1.bin"))
    assert MultiRawLoader.detect(Path(f"{root}/file1.bin+{root}/file2.bin"))

    monkeypatch.chdir(root)
    assert Path("file1.bin").exists()
    assert not MultiRawLoader.detect(Path("file1.bin"))
    assert MultiRawLoader.detect(Path("file1.bin+file2.bin"))

    fs = VirtualFilesystem()
    fs.map_file_fh("/dir/file1.bin", io.BytesIO())
    fs.map_file_fh("/dir/file2.bin", io.BytesIO())

    assert not MultiRawLoader.detect(fs.path("/dir/file1.bin"))
    assert MultiRawLoader.detect(fs.path("/dir/file1.bin+/dir/file2.bin"))



from pathlib import Path

import pytest

from dissect.target.helpers import keychain
from tests._utils import absolute_path


@pytest.fixture
def guarded_keychain():
    keychain.KEYCHAIN.clear()
    yield
    keychain.KEYCHAIN.clear()


def test_keychain_register_keychain_file(guarded_keychain):
    keychain_file = Path(absolute_path("_data/helpers/keychain/keychain.csv"))

    keychain.register_keychain_file(keychain_file)

    assert len(keychain.get_keys_without_provider()) == 1
    assert len(keychain.get_keys_for_provider("some")) == 0
    assert len(keychain.get_keys_for_provider("bitlocker")) == 2


def test_keychain_register_wildcard_value(guarded_keychain):
    keychain.register_wildcard_value("test-value")

    # Number of keys registered is equal number of supported key types, minus one for an invalid raw key
    assert len(keychain.get_keys_without_provider()) == len(keychain.KeyType) - 1

    keychain.KEYCHAIN.clear()
    keychain.register_wildcard_value("0000")

    # Valid raw key now included
    assert len(keychain.get_keys_without_provider()) == len(keychain.KeyType)

    keychain.KEYCHAIN.clear()
    keychain.register_wildcard_value("'312268-409816-583517-486695-627121-599511-664389-145640'")
    keychain.get_all_keys()[0].value == "312268-409816-583517-486695-627121-599511-664389-145640"



from dissect.target import Target
from dissect.target.loaders.overlay import Overlay2Loader


def test_overlay_loader_docker(target_linux_docker: Target) -> None:
    for container in target_linux_docker.fs.path("/var/lib/docker/image/overlay2/layerdb/mounts/").iterdir():
        assert Overlay2Loader.detect(container)
        loader = Overlay2Loader(container)
        loader.map(target_linux_docker)

    assert len(target_linux_docker.filesystems) == 4

    container_fs = target_linux_docker.filesystems[1]
    assert len(container_fs.layers) == 4
    assert len(list(container_fs.path("/").iterdir())) == 18



from __future__ import annotations

import re
from typing import Iterator

from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export

CronjobRecord = TargetRecordDescriptor(
    "unix/cronjob",
    [
        ("string", "minute"),
        ("string", "hour"),
        ("string", "day"),
        ("string", "month"),
        ("string", "weekday"),
        ("string", "user"),
        ("string", "command"),
        ("path", "source"),
    ],
)

EnvironmentVariableRecord = TargetRecordDescriptor(
    "unix/environmentvariable",
    [
        ("string", "key"),
        ("string", "value"),
        ("path", "source"),
    ],
)


class CronjobPlugin(Plugin):
    """Unix cronjob plugin."""

    def check_compatible(self) -> None:
        pass

    def parse_crontab(self, file_path) -> Iterator[CronjobRecord | EnvironmentVariableRecord]:
        for line in file_path.open("rt"):
            line = line.strip()
            if line.startswith("#") or not len(line):
                continue

            match = re.search(r"^(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(.*)$", line)
            if match:
                usr = match.group(6)
                cmd = match.group(7)
                if not str(file_path).startswith("/etc/crontab") or not str(file_path).startswith("/etc/cron.d"):
                    cmd = usr + " " + cmd
                    usr = ""

                yield CronjobRecord(
                    minute=match.group(1),
                    hour=match.group(2),
                    day=match.group(3),
                    month=match.group(4),
                    weekday=match.group(5),
                    user=usr,
                    command=cmd,
                    source=file_path,
                    _target=self.target,
                )

            s = re.search(r"^([a-zA-Z_]+[a-zA-Z[0-9_])=(.*)", line)
            if s:
                yield EnvironmentVariableRecord(
                    key=s.group(1),
                    value=s.group(2),
                    source=file_path,
                    _target=self.target,
                )

    @export(record=[CronjobRecord, EnvironmentVariableRecord])
    def cronjobs(self) -> Iterator[CronjobRecord | EnvironmentVariableRecord]:
        """Yield cronjobs on the unix system.

        A cronjob is a scheduled task/command on a Unix based system. Adversaries may use cronjobs to gain
        persistence on the system.
        """
        tabs = []
        crontab_dirs = [
            "/var/cron/tabs",
            "/var/spool/cron",
            "/var/spool/cron/crontabs",
            "/etc/cron.d",
            "/usr/local/etc/cron.d",  # FreeBSD
        ]
        for path in crontab_dirs:
            fspath = self.target.fs.path(path)
            if not fspath.exists():
                continue

            for f in fspath.iterdir():
                if not f.exists():
                    continue
                if f.is_file():
                    tabs.append(f)

        crontab_file = self.target.fs.path("/etc/crontab")
        if crontab_file.exists():
            tabs.append(crontab_file)

        crontab_file = self.target.fs.path("/etc/anacrontab")
        if crontab_file.exists():
            tabs.append(crontab_file)

        for f in tabs:
            for record in self.parse_crontab(f):
                yield record



from __future__ import annotations

from typing import Optional

from dissect.target.filesystem import Filesystem
from dissect.target.plugin import export
from dissect.target.plugins.os.unix.bsd._os import BsdPlugin
from dissect.target.target import Target


class OpenBsdPlugin(BsdPlugin):
    def __init__(self, target: Target):
        super().__init__(target)
        self._hostname_dict = self._parse_hostname_string(["/etc/myname"])

    @classmethod
    def detect(cls, target: Target) -> Optional[Filesystem]:
        for fs in target.filesystems:
            if fs.exists("/bsd") or fs.exists("/bsd.rd") or fs.exists("/bsd.mp") or fs.exists("/bsd.mp"):
                return fs

        return None

    @export(property=True)
    def version(self) -> Optional[str]:
        return None

    @export(property=True)
    def hostname(self) -> Optional[str]:
        return self._hostname_dict.get("hostname", None)



from dissect.target.plugins.child.virtuozzo import VirtuozzoChildTargetPlugin


def test_plugins_child_virtuozzo(target_unix, fs_unix):
    fs_unix.makedirs("vz/root/a")
    fs_unix.makedirs("vz/root/b")

    target_unix.add_plugin(VirtuozzoChildTargetPlugin)

    children = list(target_unix.list_children())
    assert len(children) == 2
    assert children[0].type == "virtuozzo"
    assert str(children[0].path) == "/vz/root/a"
    assert children[1].type == "virtuozzo"
    assert str(children[1].path) == "/vz/root/b"



import dataclasses
import datetime
import enum
import functools
import json
from contextlib import contextmanager
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Callable, Iterator, Optional, TextIO

import structlog

from dissect.target import Target
from dissect.target.tools.dump.utils import (
    Compression,
    Serialization,
    get_current_utc_time,
    parse_datetime_iso,
)

log = structlog.get_logger(__name__)


STATE_FILE_NAME = "target-dump.state.json"

PENDING_UPDATES_LIMIT = 10


@dataclass
class Sink:
    target_path: str
    func: str
    path: Path
    is_dirty: bool = True
    record_count: int = 0
    size_bytes: int = 0

    def __post_init__(self):
        self.func = getattr(self.func, "name", self.func)


@dataclass
class DumpState:
    target_paths: list[str]
    functions: list[str]
    serialization: str
    compression: str
    start_time: datetime.datetime
    last_update_time: datetime.datetime

    sinks: list[Sink] = dataclasses.field(default_factory=list)

    # Volatile properties
    output_dir: Optional[Path] = None
    pending_updates_count: Optional[int] = 0

    @property
    def record_count(self) -> int:
        return sum(s.record_count for s in self.sinks)

    @property
    def finished_sinks(self) -> list[Sink]:
        return [sink for sink in self.sinks if not sink.is_dirty]

    @property
    def path(self) -> Path:
        return DumpState.get_state_path(self.output_dir)

    @classmethod
    def get_state_path(cls, output_dir: Path) -> Path:
        return output_dir / STATE_FILE_NAME

    def get_full_sink_path(self, sink: Sink) -> Path:
        if not self.output_dir:
            raise ValueError("Output directory is unknown for the state")
        return self.output_dir / sink.path

    def get_sink(self, path: Path) -> Optional[Sink]:
        for sink in self.sinks:
            if sink.path == path:
                return sink

    def serialize(self) -> str:
        """Serialize state instance into a JSON formatted string"""
        state_dict = dataclasses.asdict(self)
        state_dict.pop("output_dir")
        state_dict.pop("pending_updates_count")
        serialized = json.dumps(
            state_dict,
            default=serialize_obj,
            indent=4,
            sort_keys=True,
        )
        return serialized

    def persist(self, fh: TextIO) -> None:
        """
        Write serialized state instance into profided `fh` byte stream,
        overwriting it from the beginning
        """
        fh.seek(0)
        fh.write(self.serialize())
        fh.flush()
        self.pending_updates_count = 0
        log.debug("State flushed")

    def mark_as_finished(self, target: Target, func: str) -> None:
        """
        Mark sinks that match provided `target` and `func` pair as not dirty.
        """
        matching_sinks = [
            sink for sink in self.sinks if str(sink.target_path) == str(target.path) and sink.func == func
        ]
        for sink in matching_sinks:
            sink.is_dirty = False

    def create_sink(self, sink_path: Path, stream_element) -> Sink:
        """
        Create a sink instance for provided `sink_path` and `stream_element`
        (from which `target` and `func` properties are used).
        """
        sink = Sink(
            path=sink_path,
            target_path=str(stream_element.target.path),
            func=stream_element.func,
        )
        self.sinks.append(sink)
        return sink

    def update(self, stream_element, fp_position: int) -> None:
        """
        Update a sink instance for provided `stream_element`.
        """
        sink = self.get_sink(stream_element.sink_path)

        sink.record_count += 1
        sink.size_bytes = fp_position
        self.last_update_time = get_current_utc_time()

        self.pending_updates_count += 1
        log.debug("State updated", records=self.record_count, sinks=len(self.sinks), path=self.path)

    @classmethod
    def from_dict(cls, state_dict: dict) -> "DumpState":
        """Deserialize state instance from provided dict"""
        return DumpState(
            target_paths=state_dict["target_paths"],
            functions=state_dict["functions"],
            serialization=Serialization(state_dict["serialization"]),
            compression=Compression(state_dict["compression"]),
            start_time=parse_datetime_iso(state_dict["start_time"]),
            last_update_time=parse_datetime_iso(state_dict["last_update_time"]),
            sinks=[
                Sink(
                    target_path=sink["target_path"],
                    path=Path(sink["path"]),
                    func=sink["func"],
                    record_count=sink["record_count"],
                    size_bytes=sink["size_bytes"],
                    is_dirty=sink["is_dirty"],
                )
                for sink in state_dict["sinks"]
            ],
        )

    @classmethod
    def from_path(cls, output_dir: Path) -> Optional["DumpState"]:
        """Deserialize state instance from a file in the provided output directory path"""
        state_path = DumpState.get_state_path(output_dir)
        if not state_path.exists():
            return

        with state_path.open(mode="r") as fh:
            try:
                state_dict = json.load(fh)
            except ValueError as e:
                log.warning("Can not load state from path", path=state_path, exc=e)
                return

        state = DumpState.from_dict(state_dict)
        state.output_dir = output_dir
        return state

    def get_invalid_sinks(self) -> list[Sink]:
        """Return sinks that have a mismatch between recorded size and a real file size"""
        invalid_sinks = []
        for sink in self.sinks:
            # sink file does not exist
            if not self.get_full_sink_path(sink).exists():
                invalid_sinks.append(sink)
                continue

            # recorded file size for a clean sink is incorrect
            if not sink.is_dirty and sink.size_bytes != sink.path.stat().st_size:
                invalid_sinks.append(sink)
                continue

        return invalid_sinks

    def drop_invalid_sinks(self) -> None:
        """
        Remove sinks that have a mismatch between recorded size and
        a real file size from the list of sinks.
        """
        for invalid_sink in self.get_invalid_sinks():
            self.sinks.remove(invalid_sink)
            log.debug("Ignoring invalid sink", sink=invalid_sink.path)

    def drop_dirty_sinks(self) -> None:
        """Drop sinks that are marked as "dirty" in the current state from the list of sinks"""
        dirty_sinks = [s for s in self.sinks if s.is_dirty]
        for dirty_sink in dirty_sinks:
            self.sinks.remove(dirty_sink)
            log.debug("Ignoring dirty sink", sink=dirty_sink.path)


def create_state(
    *,
    output_dir: Path,
    target_paths: list[str],
    functions: list[str],
    serialization: Serialization,
    compression: Compression = None,
) -> DumpState:
    """Create a `DumpState` instance with provided properties"""
    current_time = get_current_utc_time()
    return DumpState(
        target_paths=target_paths,
        functions=functions,
        serialization=serialization,
        compression=compression,
        start_time=current_time,
        last_update_time=current_time,
        output_dir=output_dir,
    )


@contextmanager
def persisted_state(state: DumpState) -> Iterator[Callable]:
    """Return a context manager for persisting `DumpState` instance"""

    def save_state(fh: TextIO) -> None:
        state.persist(fh)
        state.pending_updates_count = 0

    def save_state_when_over_limit(fh: TextIO) -> None:
        if state.pending_updates_count >= PENDING_UPDATES_LIMIT:
            save_state(fh)

    state.path.parent.mkdir(parents=True, exist_ok=True)

    with state.path.open(mode="w") as fh:
        try:
            yield functools.partial(save_state_when_over_limit, fh)
        finally:
            save_state(fh)


def load_state(output_dir: Path) -> Optional[DumpState]:
    """
    Load persisted `DumpState` instance from provided `output_dir` path
    and perform sink validation.
    """
    state = DumpState.from_path(output_dir)

    if not state:
        return

    # Dropping sinks that are marked as finished (clean) but the file
    # size on the disk is different from the one stored in the state object.
    state.drop_invalid_sinks()

    # Since there are no guarantees about the order of produced records
    # and the records themselves do not have unique IDs, we can not easily verify
    # what records are already stored in the dirty sink.
    # It is easier to just restart all dirty sinks.
    state.drop_dirty_sinks()

    return state


def serialize_obj(obj: Any) -> str:
    """JSON serializer for object types not serializable by `json` lib"""
    if isinstance(obj, datetime.datetime):
        return obj.isoformat()
    if isinstance(obj, Path):
        return str(obj)
    if isinstance(obj, enum.Enum):
        return obj.value
    raise TypeError("Type %s not serializable" % type(obj))



from __future__ import annotations

import logging
import urllib
from pathlib import Path
from typing import TYPE_CHECKING, Iterator, Optional, Union

from dissect.target.helpers.lazy import import_lazy
from dissect.target.helpers.loaderutil import extract_path_info

if TYPE_CHECKING:
    from dissect.target import Target

__all__ = [
    "open",
    "Loader",
    "RawLoader",
    "register",
]

log = logging.getLogger(__name__)

LOADERS: list[Loader] = []
LOADERS_BY_SCHEME: dict[str, Loader] = {}
MODULE_PATH = "dissect.target.loaders"

DirLoader: Loader = import_lazy("dissect.target.loaders.dir").DirLoader
"""A lazy loaded :class:`dissect.target.loaders.dir.DirLoader`."""

RawLoader: Loader = import_lazy("dissect.target.loaders.raw").RawLoader
"""A lazy loaded :class:`dissect.target.loaders.raw.RawLoader`."""


class Loader:
    """A base class for loading a specific path and coupling it to a :class:`Target <dissect.target.target.Target>`.

    Implementors of this class are responsible for mapping any type of source data
    to a :class:`Target <dissect.target.target.Target>`.
    Whether that's to map all VMDK files from a VMX or mapping the contents of a zip file to a virtual filesystem,
    if it's something that can be translated to a "disk", "volume" or "filesystem", you can write a loader that
    maps it into a target.

    You can do anything you want to manipulate the :class:`Target <dissect.target.target.Target>` object
    in your ``map`` function, but generally you do one of the following:

    * open a :class:`Container <dissect.target.container.Container>` and add it to ``target.disks``.
    * open a :class:`Volume <dissect.target.volume.Volume>` and add it to ``target.volumes``.
    * open a :class:`VirtualFilesystem <dissect.target.filesystem.VirtualFilesystem>`,\
    add your files into it and add it to ``target.filesystems``.

    You don't need to manually parse volumes or filesystems in your loader, just add the highest level object you have
    (e.g. a :class:`Container <dissect.target.container.Container>` of a VMDK file) to the target.
    However, sometimes you need to get creative.
    Take a look at the :class:`ITunesLoader <dissect.target.loaders.itunes.ITunesLoader>` and
    :class:`TarLoader <dissect.target.loaders.tar.TarLoader>` for some creative examples.

    Args:
        path: The target path to load.
    """

    def __init__(self, path: Path, **kwargs):
        self.path = path

    def __repr__(self):
        return f"{self.__class__.__name__}('{self.path}')"

    @staticmethod
    def detect(path: Path) -> bool:
        """Detects wether this ``Loader`` class can load this specific ``path``.

        Args:
            path: The target path to check.

        Returns:
            ``True`` if the ``path`` can be loaded by a ``Loader`` instance. ``False`` otherwise.
        """
        raise NotImplementedError()

    @staticmethod
    def find_all(path: Path, **kwargs) -> Iterator[Path]:
        """Finds all targets to load from ``path``.

        This can be used to open multiple targets from a target path that doesn't necessarily map to files on a disk.
        For example, a wildcard in a hostname a loader that opens targets from an API or Unix socket,
        such as the Carbon Black loader.

        Args:
            path: The location to a target to try and open multiple paths from.

        Returns:
            All the target paths found from the source path.
        """
        yield path

    def map(self, target: Target) -> None:
        """Maps the loaded path into a ``Target``.

        Args:
            target: The target that we're mapping into.
        """
        raise NotImplementedError()


def register(module_name: str, class_name: str, internal: bool = True) -> None:
    """Registers a ``Loader`` class inside ``LOADERS``.

    This function registers a loader using ``modname`` relative to the ``MODULE_PATH``.
    It lazily imports the module, and retrieves the specific class from it.

    Args:
        module: The module where to find the loader.
        class_name: The class to load.
        internal: Whether it is an internal module or not.
    """
    if internal:
        module = ".".join([MODULE_PATH, module_name])
    else:
        module = module_name

    loader = getattr(import_lazy(module), class_name)
    LOADERS.append(loader)
    LOADERS_BY_SCHEME[module_name] = loader


def find_loader(
    item: Path, parsed_path: Optional[urllib.parse.ParseResult] = None, fallbacks: list[Loader] = [DirLoader]
) -> Optional[Loader]:
    """Finds a :class:`Loader` class for the specific ``item``.

    This searches for a specific :class:`Loader` classs that is able to load a target pointed to by ``item``.
    Once it detects a suitable :class:`Loader` it immediately returns said :class:`Loader` class.
    It does this for all items inside the ``LOADER`` variable.

    The :class:`DirLoader <dissect.target.loaders.dir.DirLoader>` is used as the last entry
    due to how the detection methods function.

    Args:
        item: The target path to load.
        fallbacks: Fallback loaders to try.

    Returns:
        A :class:`Loader` class for the specific target if one exists.
    """
    if parsed_path:
        if loader := LOADERS_BY_SCHEME.get(parsed_path.scheme):
            return loader

    for loader in LOADERS + fallbacks:
        try:
            if loader.detect(item):
                return loader
        except ImportError as exception:
            log.info("Failed to import %s", loader)
            log.debug("", exc_info=exception)


def open(item: Union[str, Path], *args, **kwargs) -> Loader:
    """Opens a :class:`Loader` for a specific ``item``.

    This instantiates a :class:`Loader` for a specific ``item``.
    The :class:`DirLoader <dissect.target.loaders.dir.DirLoader>` is used as the last entry
    due to how the detection methods function.

    Args:
        item: The target path to load.

    Returns:
        A :class:`Loader` class for the specific target if one exists.
    """
    item, parsed_path = extract_path_info(item)

    if loader := find_loader(item, parsed_path=parsed_path):
        kwargs["parsed_path"] = parsed_path
        return loader(item, *args, **kwargs)


register("local", "LocalLoader")
register("remote", "RemoteLoader")
register("mqtt", "MQTTLoader")
register("asdf", "AsdfLoader")
register("tar", "TarLoader")
register("vmx", "VmxLoader")
register("vmwarevm", "VmwarevmLoader")
register("hyperv", "HyperVLoader")
register("pvs", "PvsLoader")
register("pvm", "PvmLoader")
register("utm", "UtmLoader")
register("ovf", "OvfLoader")
register("ova", "OvaLoader")
register("vbox", "VBoxLoader")
register("vb", "VBLoader")
register("xva", "XvaLoader")
register("vma", "VmaLoader")
register("kape", "KapeLoader")
register("tanium", "TaniumLoader")
register("itunes", "ITunesLoader")
register("ab", "AndroidBackupLoader")
register("target", "TargetLoader")
register("log", "LogLoader")
# Disabling ResLoader because of DIS-536
# register("res", "ResLoader")
register("overlay", "Overlay2Loader")
register("phobos", "PhobosLoader")
register("velociraptor", "VelociraptorLoader")
register("smb", "SmbLoader")
register("cb", "CbLoader")
register("cyber", "CyberLoader")
register("proxmox", "ProxmoxLoader")
register("multiraw", "MultiRawLoader")  # Should be last



import datetime
import json
import operator
from io import BytesIO
from typing import Iterator

import pytest
from flow.record.fieldtypes import path

from dissect.target import Target
from dissect.target.filesystem import VirtualFilesystem
from dissect.target.plugins.apps.container.docker import (
    DockerPlugin,
    convert_timestamp,
    find_installs,
    strip_log,
)
from dissect.target.plugins.os.unix._os import UnixPlugin
from tests._utils import absolute_path


@pytest.fixture
def target_linux_docker_logs(target_linux: Target, fs_linux: VirtualFilesystem) -> Iterator[Target]:
    docker_containers = absolute_path("_data/plugins/apps/container/docker/logs")
    fs_linux.map_dir("/var/lib/docker/containers", docker_containers)
    yield target_linux


def test_docker_plugin_data_roots(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    fs_unix.makedirs("var/lib/docker")
    fs_unix.makedirs("tmp/foo/bar")
    fs_unix.makedirs("tmp/another/docker")
    fs_unix.map_file_fh("/etc/docker/daemon.json", BytesIO(b'{"data-root": "/tmp/foo/bar"}'))
    fs_unix.map_file_fh("/root/.docker/daemon.json", BytesIO(b'{"data-root": "/tmp/another/docker"}'))

    assert [str(p) for p in find_installs(target_unix_users)] == [
        "/var/lib/docker",
        "/tmp/foo/bar",
        "/tmp/another/docker",
    ]


def test_docker_plugin_images(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    fs_unix.map_file(
        "/var/lib/docker/image/overlay2/repositories.json",
        absolute_path("_data/plugins/apps/container/docker/repositories.json"),
    )

    hash = "6b7dfa7e8fdbe18ad425dd965a1049d984f31cf0ad57fa6d5377cca355e65f03"
    fs_unix.map_file(
        f"/var/lib/docker/image/overlay2/imagedb/content/sha256/{hash}",
        absolute_path("_data/plugins/apps/container/docker/image_metadata.json"),
    )

    target_unix_users.add_plugin(DockerPlugin)
    results = list(target_unix_users.docker.images())
    assert len(results) == 2
    assert results[0].tag == "ubuntu:latest"
    assert results[0].hash == f"sha256:{hash}"
    assert results[0].image_id == "6b7dfa7e8fdb"
    assert results[0].created == datetime.datetime(2022, 12, 9, 1, 20, 31, 321639, tzinfo=datetime.timezone.utc)


def test_docker_plugin_containers(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    id = "d3adb33fd3adb33fd3adb33fd3adb33fd3adb33fd3adb33fd3adb33fd3adb33f"

    fs_unix.map_file(
        f"/var/lib/docker/containers/{id}/config.v2.json",
        absolute_path("_data/plugins/apps/container/docker/container_running.json"),
    )
    target_unix_users.add_plugin(DockerPlugin)
    results = list(target_unix_users.docker.containers())

    assert len(results) == 1

    result = results[0]

    assert result.container_id == id
    assert result.image == "exampleimage:1.33.7"
    assert result.image_id == "d3adb33fd3adb33fd3adb33fd3adb33fd3adb33fd3adb33fd3adb33fd3adb33f"
    assert result.created == datetime.datetime(2022, 12, 19, 13, 37, 1, 247519, tzinfo=datetime.timezone.utc)
    assert bool(result.running) is True
    assert result.started == datetime.datetime(2022, 12, 19, 13, 37, 1, 247519, tzinfo=datetime.timezone.utc)
    assert result.finished == datetime.datetime(1, 1, 1, 00, 00, 00, 000000, tzinfo=datetime.timezone.utc)
    assert result.ports == str({"1234/tcp": "0.0.0.0:1234", "5678/tcp": "0.0.0.0:5678"})
    assert result.names == "example_container_name"
    assert result.volumes == ["/tmp/test:/test"]
    assert result.config_path == path.from_posix(f"/var/lib/docker/containers/{id}/config.v2.json")
    assert result.mount_path == path.from_posix(f"/var/lib/docker/image/overlay2/layerdb/mounts/{id}")


def test_docker_plugin_logs(target_linux_docker_logs: Target) -> None:
    target_linux_docker_logs.add_plugin(DockerPlugin)
    results = list(target_linux_docker_logs.docker.logs())
    results.sort(key=operator.attrgetter("ts"))

    assert len(results) == 288

    # json log driver
    assert results[0].ts == datetime.datetime(2023, 11, 9, 8, 42, 57, 79838, tzinfo=datetime.timezone.utc)
    assert results[0].container == "f988f88e221d97930a665712cf16ab520f7e2c5af395660c145df93aebedf071"
    assert results[0].stream == "stdout"
    assert results[0].message == "/ # "

    # local log driver (protobuf)
    assert results[-1].ts == datetime.datetime(2023, 11, 9, 9, 52, 52, 587579, tzinfo=datetime.timezone.utc)
    assert results[-1].container == "0627aa2d32de2478f4a3e8bb3c655ea7baa1a3463d8cee41263655244fe4717c"
    assert results[-1].stream == "stdout"
    assert results[-1].message == "exit"


def test_docker_plugin_logs_raw(target_linux_docker_logs: Target) -> None:
    target_linux_docker_logs.add_plugin(DockerPlugin)
    results = list(target_linux_docker_logs.docker.logs(raw_messages=True))
    results.sort(key=operator.attrgetter("ts"))

    assert len(results) == 288
    assert results[0].message == "/ # \x1b[6n\r\n"
    assert results[-1].message == "\x1b[?2004l\rexit\r"


def test_docker_plugin_timestamps() -> None:
    # Should not alter already correct timestamps
    assert convert_timestamp("2022-12-19T13:37:00.123456") == "2022-12-19T13:37:00.123456"
    assert convert_timestamp("2022-12-19T13:37:00.123456Z") == "2022-12-19T13:37:00.123456Z"

    # Should convert nanosecond timestamps to microsecond timestamps
    assert convert_timestamp("2022-12-19T13:37:00.123456789Z") == "2022-12-19T13:37:00.123456Z"
    assert convert_timestamp("2022-12-19T13:37:00.12345678Z") == "2022-12-19T13:37:00.123456Z"
    assert convert_timestamp("2022-12-19T13:37:00.123456789+01:00") == "2022-12-19T13:37:00.123456+01:00"


def test_backspace_interpretation() -> None:
    # Should 'execute' backspaces found in input
    input = (
        '~ # \x1b[6necho \'\x08\x1b[J"ths \x08\x1b[J\x08\x1b[Js \x08\x1b[J\x08\x1b[Jis is a secret!" > secret.txt\r\n'
    )
    assert strip_log(input, exc_backspace=True) == '~ # echo "this is a secret!" > secret.txt'


def test_regression_running_container_parsing(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    """test if we correctly discover and reconstruct exposed container ports and commands on a running container"""

    id = "deadbeef"
    config = {
        "ID": id,
        "Image": "sha256:blabla",
        "Name": "/foo",
        "Path": "/bin/some-binary",
        "Driver": "overlay2",
        "Args": [
            "--some-argument 1",
            "--another-argument 2",
        ],
        "State": {
            "Running": True,
            "Pid": 1337,
            "StartedAt": "2024-12-31T13:37:00.123456789Z",
        },
        "Config": {
            "Hostname": "foo",
            "ExposedPorts": {
                "1337/tcp": {},
                "1337/udp": {},
            },
            "Image": "docker.io/debian",
        },
        "NetworkSettings": {
            "Ports": {},
        },
        "MountPoints": {
            "/dest/file.txt": {
                "Source": "/somewhere/on/host/file.txt",
                "Destination": "/dest/file.txt",
            },
        },
    }

    fs_unix.map_file_fh(f"/var/lib/docker/containers/{id}/config.v2.json", BytesIO(json.dumps(config).encode()))
    fs_unix.map_file_fh("/etc/hostname", BytesIO(b"hostname"))

    target_unix.add_plugin(UnixPlugin)
    target_unix.add_plugin(DockerPlugin)

    results = list(target_unix.docker.containers())
    assert len(results) == 1

    assert results[0].hostname == "hostname"
    assert results[0].container_id == id
    assert results[0].image == "docker.io/debian"
    assert results[0].image_id == "blabla"
    assert results[0].command == "/bin/some-binary --some-argument 1 --another-argument 2"
    assert results[0].created is None
    assert results[0].running
    assert results[0].pid == 1337
    assert results[0].started == datetime.datetime(2024, 12, 31, 13, 37, 0, 123456, tzinfo=datetime.timezone.utc)
    assert results[0].finished is None
    assert results[0].ports == str({"1337/tcp": "0.0.0.0:1337", "1337/udp": "0.0.0.0:1337"})
    assert results[0].names == "foo"
    assert results[0].volumes == ["/somewhere/on/host/file.txt:/dest/file.txt"]
    assert results[0].mount_path == "/var/lib/docker/image/overlay2/layerdb/mounts/deadbeef"
    assert results[0].config_path == "/var/lib/docker/containers/deadbeef/config.v2.json"



from unittest.mock import patch

from dissect.target.loaders.tanium import TaniumLoader
from tests._utils import absolute_path, mkdirs


@patch("dissect.target.filesystems.dir.DirectoryFilesystem.ntfs", None, create=True)
def test_tanium_loader(target_bare, tmp_path):
    root = tmp_path
    mkdirs(root, ["file/C/windows/system32", "file/C/$Extend", "file/D/test", "file/E/test"])

    # Only need this to exist up until the root directory record to make dissect.ntfs happy
    with open(absolute_path("_data/plugins/filesystem/ntfs/mft/mft.raw"), "rb") as fh:
        (root / "file/C/$MFT").write_bytes(fh.read(10 * 1025))

    # Add one record so we can test if it works
    data = bytes.fromhex(
        "5800000002000000c100000000000100bf000000000001002003010000000000"
        "6252641a86a4d7010381008000000000000000002000000018003c0069007300"
        "2d00310035005000320036002e0074006d00700000000000"
    )
    (root / "file/C/$Extend/$UsnJrnl_$J").write_bytes(data)

    assert TaniumLoader.detect(root)

    loader = TaniumLoader(root)
    loader.map(target_bare)
    target_bare.apply()

    assert "sysvol" in target_bare.fs.mounts
    assert "c:" in target_bare.fs.mounts
    assert "d:" in target_bare.fs.mounts
    assert "e:" in target_bare.fs.mounts

    # The 3 found drive letter directories + the fake NTFS filesystem
    assert len(target_bare.filesystems) == 4
    # The 3 found drive letters + sysvol + the fake NTFS filesystem at /$fs$
    assert len(target_bare.fs.mounts) == 5
    assert len(list(target_bare.fs.mounts["c:"].ntfs.usnjrnl.records())) == 1



from dissect.target.filesystem import VirtualFile, VirtualFilesystem
from dissect.target.plugins.filesystem.walkfs import WalkFSPlugin
from dissect.target.target import Target


def test_walkfs_plugin(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    fs_unix.map_file_entry("/path/to/some/file", VirtualFile(fs_unix, "file", None))
    fs_unix.map_file_entry("/path/to/some/other/file.ext", VirtualFile(fs_unix, "file.ext", None))
    fs_unix.map_file_entry("/root_file", VirtualFile(fs_unix, "root_file", None))
    fs_unix.map_file_entry("/other_root_file.ext", VirtualFile(fs_unix, "other_root_file.ext", None))
    fs_unix.map_file_entry("/.test/test.txt", VirtualFile(fs_unix, "test.txt", None))
    fs_unix.map_file_entry("/.test/.more.test.txt", VirtualFile(fs_unix, ".more.test.txt", None))

    target_unix.add_plugin(WalkFSPlugin)

    results = list(target_unix.walkfs())
    assert len(results) == 14
    assert sorted([r.path for r in results]) == [
        "/",
        "/.test",
        "/.test/.more.test.txt",
        "/.test/test.txt",
        "/etc",
        "/other_root_file.ext",
        "/path",
        "/path/to",
        "/path/to/some",
        "/path/to/some/file",
        "/path/to/some/other",
        "/path/to/some/other/file.ext",
        "/root_file",
        "/var",
    ]



import sys
from pathlib import Path
from typing import Iterator
from unittest.mock import call, patch

import pytest

from dissect.target.plugin import PLUGINS, find_py_files, load_modules_from_paths


@pytest.fixture
def environment_path(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> Iterator[Path]:
    monkeypatch.setenv("DISSECT_PLUGINS", str(tmp_path.absolute()))
    yield tmp_path


def registry_file(path: str) -> Path:
    return Path(__file__).parent / "_data/registration/" / path


def copy_different_plugin_files(path: Path, file_name: str) -> None:
    plugin_file = path / file_name
    plugin_file.touch()
    plugin = registry_file(file_name)
    plugin_file.write_text(plugin.read_text())


def test_load_environment_variable_empty_string() -> None:
    with patch("dissect.target.plugin.find_py_files") as mocked_find_py_files:
        load_modules_from_paths([])
        mocked_find_py_files.assert_not_called()


def test_load_environment_variable_comma_seperated_string() -> None:
    with patch("dissect.target.plugin.find_py_files") as mocked_find_py_files:
        load_modules_from_paths([Path(""), Path("")])
        mocked_find_py_files.assert_has_calls(calls=[call(Path(""))])


def test_filter_file(tmp_path: Path) -> None:
    file = tmp_path / "hello.py"
    file.touch()

    assert list(find_py_files(file)) == [file]

    test_file = tmp_path / "non_existent_file"
    assert list(find_py_files(test_file)) == []

    test_file = tmp_path / "__init__.py"
    test_file.touch()
    assert list(find_py_files(test_file)) == []


@pytest.mark.parametrize(
    "filename, empty_list",
    [
        ("__init__.py", True),
        ("__pycache__/help.pyc", True),
        ("hello/test.py", False),
    ],
)
def test_filter_directory(tmp_path: Path, filename: str, empty_list: bool) -> None:
    file = tmp_path / filename
    file.parent.mkdir(parents=True, exist_ok=True)
    file.touch()

    if empty_list:
        assert list(find_py_files(tmp_path)) == []
    else:
        assert file in list(find_py_files(tmp_path))


def test_new_plugin_registration(environment_path: Path) -> None:
    copy_different_plugin_files(environment_path, "plugin.py")
    load_modules_from_paths([environment_path])

    assert "plugin" in PLUGINS


def test_loader_registration(environment_path: Path) -> None:
    with patch("dissect.target.loader.LOADERS", []) as mocked_loaders, patch(
        "dissect.target.loader.LOADERS_BY_SCHEME", {}
    ):
        copy_different_plugin_files(environment_path, "loader.py")
        load_modules_from_paths([environment_path])

        assert len(mocked_loaders) == 1
        assert mocked_loaders[0].__name__ == "TestLoader"


def test_filesystem_registration(environment_path: Path) -> None:
    with patch("dissect.target.filesystem.FILESYSTEMS", []) as mocked_filesystems:
        copy_different_plugin_files(environment_path, "filesystem.py")
        load_modules_from_paths([environment_path])

        assert len(mocked_filesystems) == 1
        assert mocked_filesystems[0].__name__ == "TestFilesystem"


def test_container_registration(environment_path: Path) -> None:
    with patch("dissect.target.container.CONTAINERS", []) as mocked_containers:
        copy_different_plugin_files(environment_path, "container.py")
        load_modules_from_paths([environment_path])

        assert len(mocked_containers) == 1
        assert mocked_containers[0].__name__ == "TestContainer"


def test_register_file(environment_path: Path) -> None:
    with patch("dissect.target.container.CONTAINERS", []) as mocked_containers:
        copy_different_plugin_files(environment_path, "container.py")
        load_modules_from_paths([environment_path / "container.py"])

        assert len(mocked_containers) == 1
        assert mocked_containers[0].__name__ == "TestContainer"


@pytest.mark.parametrize(
    "filename, expected_module",
    [
        ("test.py", "test"),
        ("hello_world/help.py", "hello_world.help"),
        ("path/to/file.py", "path.to.file"),
    ],
)
def test_filesystem_module_registration(environment_path: Path, filename: str, expected_module: str) -> None:
    path = environment_path / filename
    path.parent.mkdir(parents=True, exist_ok=True)
    path.touch()

    load_modules_from_paths([environment_path])

    assert expected_module in sys.modules.keys()



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from dissect.hypervisor import pvs

from dissect.target.containers.hdd import HddContainer
from dissect.target.loader import Loader

if TYPE_CHECKING:
    from dissect.target import Target


class PvsLoader(Loader):
    """Parallels VM configuration file (config.pvs)."""

    def __init__(self, path: Path, **kwargs):
        path = path.resolve()

        super().__init__(path)
        self.pvs = pvs.PVS(path.open("rt"))
        self.base_dir = path.parent

    @staticmethod
    def detect(path: Path) -> bool:
        return path.suffix.lower() == ".pvs"

    def map(self, target: Target) -> None:
        for disk in self.pvs.disks():
            path = self.base_dir.joinpath(disk)
            try:
                target.disks.add(HddContainer(path))
            except Exception:
                target.log.exception("Failed to load HDD: %s", disk)



from unittest.mock import call, patch

from dissect.target.loaders.vbox import VBoxLoader


@patch("pathlib.Path")
@patch("dissect.target.loaders.vbox.VBox")
def test_vbox_loader(VBox, Path, target_bare) -> None:
    VBox.return_value = VBox
    VBox.disks.return_value = [Path("/mock.vdi")]
    vbox_loader = VBoxLoader(Path("/mock.vbox"))
    vbox_loader.map(target_bare)
    expected = [call.disks()]
    del VBox.mock_calls[0]
    assert expected == VBox.mock_calls
    assert len(target_bare.disks) == 1



from __future__ import annotations

from typing import Any, Iterator, Union

from flow.record.fieldtypes.net import IPAddress, IPNetwork
from flow.record.fieldtypes.net.ipv4 import Address, addr_long, addr_str, mask_to_bits

from dissect.target.helpers.record import (
    MacInterfaceRecord,
    UnixInterfaceRecord,
    WindowsInterfaceRecord,
)
from dissect.target.plugin import Plugin, export, internal
from dissect.target.target import Target

InterfaceRecord = Union[UnixInterfaceRecord, WindowsInterfaceRecord, MacInterfaceRecord]


class NetworkPlugin(Plugin):
    """Generic implementation for network interfaces plugin."""

    __namespace__ = "network"

    def __init__(self, target: Target):
        super().__init__(target)
        self._interface_list: list[InterfaceRecord] | None = None

    def check_compatible(self) -> None:
        pass

    def _interfaces(self) -> Iterator[InterfaceRecord]:
        yield from ()

    def _get_record_type(self, field_name: str) -> Iterator[Any]:
        for record in self.interfaces():
            if (output := getattr(record, field_name, None)) is None:
                continue

            if isinstance(output, list):
                yield from output
            else:
                yield output

    @export(record=InterfaceRecord)
    def interfaces(self) -> Iterator[InterfaceRecord]:
        """Yield interfaces."""
        # Only search for the interfaces once
        if self._interface_list is None:
            self._interface_list = list(self._interfaces())

        yield from self._interface_list

    @export
    def ips(self) -> list[IPAddress]:
        """Return IP addresses as list of :class:`IPAddress`."""
        return list(set(self._get_record_type("ip")))

    @export
    def gateways(self) -> list[IPAddress]:
        """Return gateways as list of :class:`IPAddress`."""
        return list(set(self._get_record_type("gateway")))

    @export
    def macs(self) -> list[str]:
        """Return MAC addresses as list of :class:`str`."""
        return list(set(self._get_record_type("mac")))

    @export
    def dns(self) -> list[str | IPAddress]:
        """Return DNS addresses as list of :class:`str`."""
        return list(set(self._get_record_type("dns")))

    @internal
    def with_ip(self, ip_addr: str) -> Iterator[InterfaceRecord]:
        for interface in self.interfaces():
            if ip_addr in interface.ip:
                yield interface

    @internal
    def with_mac(self, mac: str) -> Iterator[InterfaceRecord]:
        for interface in self.interfaces():
            if mac in interface.mac:
                yield interface

    @internal
    def in_cidr(self, cidr: str) -> Iterator[InterfaceRecord]:
        cidr = IPNetwork(cidr)
        for interface in self.interfaces():
            if any(ip_addr in cidr for ip_addr in interface.ip):
                yield interface

    def calculate_network(self, ips: int | Address, subnets: int | Address) -> Iterator[str]:
        for ip, subnet_mask in zip(ips, subnets):
            subnet_mask_int = addr_long(subnet_mask)
            cidr = mask_to_bits(subnet_mask_int)
            network_address = addr_str(addr_long(ip) & subnet_mask_int)
            yield f"{network_address}/{cidr}"



import logging
import pathlib
from typing import BinaryIO, Iterator, Optional, Union

from dissect.fve import luks
from dissect.util.stream import AlignedStream

from dissect.target.exceptions import VolumeSystemError
from dissect.target.helpers.keychain import KeyType
from dissect.target.volume import EncryptedVolumeSystem, Volume

log = logging.getLogger(__name__)


class LUKSVolumeSystemError(VolumeSystemError):
    pass


class LUKSVolumeSystem(EncryptedVolumeSystem):
    __type__ = "luks"

    def __init__(self, fh: Union[BinaryIO, list[BinaryIO]], *args, **kwargs):
        super().__init__(fh, *args, **kwargs)
        self.luks = luks.LUKS(fh)

    @staticmethod
    def _detect(fh: BinaryIO) -> bool:
        return luks.is_luks_volume(fh)

    def _volumes(self) -> Iterator[Volume]:
        if isinstance(self.fh, Volume):
            volume_details = dict(
                number=self.fh.number,
                offset=self.fh.offset,
                vtype=self.fh.type,
                name=self.fh.name,
            )
        else:
            volume_details = dict(
                number=1,
                offset=0,
                vtype=None,
                name=None,
            )

        stream = self.unlock_volume()
        yield Volume(
            fh=stream,
            size=stream.size,
            raw=self.fh,
            vs=self,
            **volume_details,
        )

    def unlock_with_volume_encryption_key(
        self, key: bytes, keyslot: Optional[int] = None, is_wildcard: bool = False
    ) -> None:
        try:
            if keyslot is None:
                for keyslot in self.luks.keyslots.keys():
                    try:
                        self.luks.unlock(key, keyslot)
                        break
                    except ValueError:
                        continue
                else:
                    raise ValueError("Failed to find matching keyslot for provided volume encryption key")
            else:
                self.luks.unlock(key, keyslot)

            log.debug("Unlocked LUKS volume with provided volume encryption key")
        except ValueError:
            if not is_wildcard:
                log.exception("Failed to unlock LUKS volume with provided volume encryption key")

    def unlock_with_passphrase(self, passphrase: str, keyslot: Optional[int] = None, is_wildcard: bool = False) -> None:
        try:
            self.luks.unlock_with_passphrase(passphrase, keyslot)
            log.debug("Unlocked LUKS volume with provided passphrase")
        except ValueError:
            if not is_wildcard:
                log.exception("Failed to unlock LUKS volume with provided passphrase")

    def unlock_with_key_file(
        self, key_file: pathlib.Path, keyslot: Optional[int] = None, is_wildcard: bool = False
    ) -> None:
        if not key_file.exists():
            if not is_wildcard:
                log.error("Provided key file does not exist: %s", key_file)
            return

        try:
            self.luks.unlock_with_key_file(key_file, keyslot=keyslot)
            log.debug("Unlocked LUKS volume with key file %s", key_file)
        except ValueError:
            if not is_wildcard:
                log.exception("Failed to unlock LUKS volume with key file %s", key_file)

    def unlock_volume(self) -> AlignedStream:
        keyslots = list(map(str, self.luks.keyslots.keys()))
        keys = self.get_keys_for_identifiers(keyslots) + self.get_keys_without_identifier()

        for key in keys:
            try:
                keyslot = int(key.identifier)
            except Exception:
                keyslot = None

            if key.key_type == KeyType.RAW:
                self.unlock_with_volume_encryption_key(key.value, keyslot, key.is_wildcard)
            if key.key_type == KeyType.PASSPHRASE:
                self.unlock_with_passphrase(key.value, keyslot, key.is_wildcard)
            elif key.key_type == KeyType.FILE:
                key_file = pathlib.Path(key.value)
                self.unlock_with_key_file(key_file, keyslot, key.is_wildcard)

            if self.luks.unlocked:
                log.info("Volume %s unlocked with %s (keyslot: %d)", self.fh, key, self.luks._active_keyslot_id)
                break

        if self.luks.unlocked:
            return self.luks.open()
        else:
            raise LUKSVolumeSystemError("Failed to unlock LUKS volume")



import logging
from collections import defaultdict
from typing import BinaryIO, Iterator, Union

from dissect.vmfs import lvm
from dissect.vmfs.c_vmfs import c_vmfs

from dissect.target.volume import LogicalVolumeSystem, Volume

log = logging.getLogger(__name__)


class VmfsVolumeSystem(LogicalVolumeSystem):
    __type__ = "vmfs"

    def __init__(self, fh: Union[BinaryIO, list[BinaryIO]], *args, **kwargs):
        self.lvm = lvm.LVM(fh)
        super().__init__(fh, *args, **kwargs)

    @classmethod
    def open_all(cls, volumes: list[BinaryIO]) -> Iterator[LogicalVolumeSystem]:
        lvm_extents = defaultdict(list)

        for vol in volumes:
            if not cls.detect_volume(vol):
                continue

            extent = lvm.Extent(vol)
            lvm_extents[extent.uuid].append(extent)

        for pvs in lvm_extents.values():
            try:
                yield cls(pvs)
            except Exception:
                continue

    @staticmethod
    def _detect(fh: BinaryIO) -> bool:
        return VmfsVolumeSystem.detect_volume(fh)

    @staticmethod
    def _detect_volume(fh: BinaryIO) -> bool:
        fh.seek(c_vmfs.VMFS_LVM_DEVICE_META_BASE)
        sector = fh.read(512)
        return int.from_bytes(sector[:4], "little") == c_vmfs.VMFS_LVM_DEVICE_META_MAGIC

    def _volumes(self) -> Iterator[Volume]:
        try:
            name = next(extent.fh.name for extent in self.fh if hasattr(extent.fh, "name"))
        except StopIteration:
            name = None

        yield Volume(
            self.lvm,
            1,
            None,
            self.lvm.size,
            None,
            name,
            self.lvm.uuid,
            raw=self.lvm,
            vs=self,
        )



import logging
from functools import partial
from typing import Callable, Iterable, Iterator

from flow.record import GroupedRecord, Record, RecordDescriptor, fieldtypes

from dissect.target import Target
from dissect.target.exceptions import FileNotFoundError, FilesystemError
from dissect.target.helpers.fsutil import TargetPath
from dissect.target.helpers.hashutil import common
from dissect.target.helpers.utils import StrEnum

__all__ = ("get_modifier_function", "Modifier", "ModifierFunc")

RECORD_NAME = "filesystem/file/digest"
NAME_SUFFIXES = ["_resolved", "_digest"]
RECORD_TYPES = ["path", "digest"]

ModifierFunc = Callable[[Target, Record], GroupedRecord]


class Modifier(StrEnum):
    RESOLVE = "resolve"
    HASH = "hash"


def _create_modified_record(
    record_name: str, field_name: str, field_info: Iterable[tuple[str, str, TargetPath]]
) -> Record:
    record_kwargs = dict()
    record_def = list()
    for type, name_suffix, data in field_info:
        extended_field_name = f"{field_name}{name_suffix}"
        record_kwargs.update({extended_field_name: data})
        record_def.append((type, extended_field_name))

    _record = RecordDescriptor(record_name, record_def)
    return _record(**record_kwargs)


def _resolve_path_records(field_name: str, resolved_path: TargetPath) -> Record:
    """Resolve files from path fields inside the record."""
    type_info = [("path", "_resolved", resolved_path)]
    return _create_modified_record("filesystem/file/resolved", field_name, type_info)


def _hash_path_records(field_name: str, resolved_path: TargetPath) -> Record:
    """Hash files from path fields inside the record.

    Args:
        field_name: Name of the field.
        resolved_path: Path to the file we should hash.

    Raises:
        FileNotFoundError: Raised if the provided ``resolved_path`` does not exist or is not a file on the target.

    Returns: Modified record with digests of path field types.
    """

    if not resolved_path.exists() or not resolved_path.is_file():
        raise FileNotFoundError(f"Path not found or is not a file: '{resolved_path}'")

    with resolved_path.open() as fh:
        path_hash = common(fh)

    type_info = zip(RECORD_TYPES, NAME_SUFFIXES, [resolved_path, path_hash])

    return _create_modified_record("filesystem/file/digest", field_name, type_info)


MODIFIER_MAPPING = {
    Modifier.RESOLVE: _resolve_path_records,
    Modifier.HASH: _hash_path_records,
}


def _resolve_path_types(target: Target, record: Record) -> Iterator[tuple[str, TargetPath]]:
    for field_name, field_type in record._field_types.items():
        if not issubclass(field_type, (fieldtypes.path, fieldtypes.command)):
            continue

        path = getattr(record, field_name, None)
        if path is None:
            continue

        if isinstance(path, fieldtypes.command):
            path = path.executable

        yield field_name, target.resolve(str(path))


def modify_record(target: Target, record: Record, modifier_function: ModifierFunc) -> GroupedRecord:
    additional_records = []

    for field_name, resolved_path in _resolve_path_types(target, record):
        try:
            _record = modifier_function(field_name, resolved_path)
        except FilesystemError as e:
            level = logging.INFO if isinstance(e, FileNotFoundError) else logging.WARNING
            target.log.log(
                level,
                "Unable to modify record '%s' with function '%s': %s",
                record._desc.name,
                modifier_function.__name__,
                e,
            )
            target.log.debug("", exc_info=e)

        else:
            additional_records.append(_record)

    if not additional_records:
        return record

    return GroupedRecord(record._desc.name, [record] + additional_records)


def _noop(_target: Target, record: Record):
    return record


def get_modifier_function(modifier_type: Modifier) -> ModifierFunc:
    if func := MODIFIER_MAPPING.get(modifier_type):
        return partial(modify_record, modifier_function=func)

    return _noop



import tarfile
from pathlib import Path
from unittest.mock import MagicMock, call, patch

from dissect.target.loaders.ova import OvaLoader
from dissect.target.target import Target


@patch("dissect.target.loaders.ovf.container")
@patch("dissect.target.loaders.ovf.ovf.OVF")
def test_ova_loader(OVF: MagicMock, container: MagicMock, target_bare: Target, tmp_path: Path):
    tf = tarfile.open((tmp_path / "test.ova"), "w")
    tf.addfile(tarfile.TarInfo("test.ovf"), b"")
    tf.addfile(tarfile.TarInfo("disk.vmdk"), b"")
    tf.close()

    OVF.return_value = OVF
    OVF.disks.return_value = ["disk.vmdk"]
    container.open.return_value = MagicMock()

    ova_loader = OvaLoader(tmp_path / "test.ova")
    ova_loader.map(target_bare)

    assert len(target_bare.disks) == 1
    assert container.open.mock_calls == [call(ova_loader.base_dir / "disk.vmdk")]



import re
from typing import Iterator, Optional

from dissect.target.helpers.fsutil import TargetPath
from dissect.target.helpers.record import UnixUserRecord
from dissect.target.helpers.utils import year_rollover_helper
from dissect.target.plugin import export
from dissect.target.plugins.os.unix.history import (
    CommandHistoryPlugin,
    CommandHistoryRecord,
)

RE_CITRIX_NETSCALER_BASH_HISTORY_DATE = re.compile(r"(?P<date>[^<]+)\s")

CITRIX_NETSCALER_BASH_HISTORY_RE = re.compile(
    r"""
        (?P<date>[^<]+)
        \s
        <
            (?P<syslog_facility>[^\.]+)
            \.
            (?P<syslog_loglevel>[^>]+)
        >
        \s
        (?P<hostname>[^\s]+)
        \s
        (?P<process_name>[^\[]+)
            \[
                (?P<process_id>\d+)
            \]
        :
        \s
        (?P<username>.*)\s
        on\s
            (?P<destination>[^\s]+)\s
        shell_command=
        \"
            (?P<command>.*)
        \"
    $
    """,
    re.VERBOSE,
)


class CitrixCommandHistoryPlugin(CommandHistoryPlugin):
    """Citrix command history plugin."""

    COMMAND_HISTORY_ABSOLUTE_PATHS = (("citrix-netscaler-bash", "/var/log/bash.log*"),)
    COMMAND_HISTORY_RELATIVE_PATHS = CommandHistoryPlugin.COMMAND_HISTORY_RELATIVE_PATHS + (
        ("citrix-netscaler-cli", ".nscli_history"),
    )

    def _find_history_files(self) -> list[tuple[str, TargetPath, Optional[UnixUserRecord]]]:
        """Find history files on the target that this plugin can parse."""
        history_files = []
        for shell, history_absolute_path_glob in self.COMMAND_HISTORY_ABSOLUTE_PATHS:
            for path in self.target.fs.path("/").glob(history_absolute_path_glob.lstrip("/")):
                history_files.append((shell, path, None))

        # Also utilize the _find_history_files function of the parent class
        history_files.extend(super()._find_history_files())
        return history_files

    def _find_user_by_name(self, username: str) -> Optional[UnixUserRecord]:
        """Cached function to return the matching UnixUserRecord for a given username."""
        if username is None:
            return None

        user_details = self.target.user_details.find(username=username)
        return user_details.user if user_details else None

    @export(record=CommandHistoryRecord)
    def commandhistory(self) -> Iterator[CommandHistoryRecord]:
        """Return shell history for all users.

        When using a shell, history of the used commands is kept on the system.
        """

        for shell, history_path, user in self._history_files:
            if shell == "citrix-netscaler-cli":
                yield from self.parse_netscaler_cli_history(history_path, user)
            elif shell == "citrix-netscaler-bash":
                yield from self.parse_netscaler_bash_history(history_path)

    def parse_netscaler_bash_history(self, path: TargetPath) -> Iterator[CommandHistoryRecord]:
        """Parse bash.log* contents."""
        for ts, line in year_rollover_helper(path, RE_CITRIX_NETSCALER_BASH_HISTORY_DATE, "%b %d %H:%M:%S "):
            line = line.strip()
            if not line:
                continue

            match = CITRIX_NETSCALER_BASH_HISTORY_RE.match(line)
            if not match:
                continue

            group = match.groupdict()
            command = group.get("command")
            user = self._find_user_by_name(group.get("username"))

            yield CommandHistoryRecord(
                ts=ts,
                command=command,
                shell="citrix-netscaler-bash",
                source=path,
                _target=self.target,
                _user=user,
            )

    def parse_netscaler_cli_history(
        self, history_file: TargetPath, user: UnixUserRecord
    ) -> Iterator[CommandHistoryRecord]:
        """Parses the history file of the Citrix Netscaler CLI.

        The only difference compared to generic bash history files is that the first line will start with
        ``_HiStOrY_V2_``, which we will skip.
        """
        for idx, line in enumerate(history_file.open("rt")):
            if not (line := line.strip()):
                continue

            if idx == 0 and line == "_HiStOrY_V2_":
                continue

            yield CommandHistoryRecord(
                ts=None,
                command=line,
                shell="citrix-netscaler-cli",
                source=history_file,
                _target=self.target,
                _user=user,
            )



from __future__ import annotations

import itertools
import logging
import re
from abc import ABC, abstractmethod
from datetime import datetime
from functools import lru_cache
from itertools import chain
from pathlib import Path
from typing import Any, Iterator

from dissect.target import Target
from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.fsutil import open_decompress
from dissect.target.helpers.record import DynamicDescriptor, TargetRecordDescriptor
from dissect.target.helpers.utils import year_rollover_helper
from dissect.target.plugin import Plugin, alias, export

log = logging.getLogger(__name__)

RE_TS = re.compile(r"^[A-Za-z]{3}\s*\d{1,2}\s\d{1,2}:\d{2}:\d{2}")
RE_TS_ISO = re.compile(r"^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\.\d{6}\+\d{2}:\d{2}")
RE_LINE = re.compile(
    r"""
    \d{2}:\d{2}\s                           # First match on the similar ending of the different timestamps
    (?P<hostname>\S+)\s                     # The hostname
    (?P<service>\S+?)(\[(?P<pid>\d+)\])?:   # The service with optionally the PID between brackets
    \s*(?P<message>.+?)\s*$                 # The log message stripped from spaces left and right
    """,
    re.VERBOSE,
)

# Generic regular expressions
RE_IPV4_ADDRESS = re.compile(
    r"""
    ((25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}   # First three octets
    (25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)          # Last octet
    """,
    re.VERBOSE,
)
RE_USER = re.compile(r"for ([^\s]+)")


class BaseService(ABC):
    @classmethod
    @abstractmethod
    def parse(cls, message: str) -> dict[str, any]:
        pass


class SudoService(BaseService):
    """Parsing of sudo service messages in the auth log."""

    RE_SUDO_COMMAND = re.compile(
        r"""
        TTY=(?P<tty>\w+\/\w+)\s;\s          # The TTY -> TTY=pts/0 ;
        PWD=(?P<pwd>[\/\w]+)\s;\s           # The current working directory -> PWD="/home/user" ;
        USER=(?P<effective_user>\w+)\s;\s   # The effective user -> USER=root ;
        COMMAND=(?P<command>.+)$            # The command -> COMMAND=/usr/bin/whoami
        """,
        re.VERBOSE,
    )

    @classmethod
    def parse(cls, message: str) -> dict[str, str]:
        """Parse auth log message from sudo."""
        if not (match := cls.RE_SUDO_COMMAND.search(message)):
            return {}

        additional_fields = {}
        for key, value in match.groupdict().items():
            additional_fields[key] = value

        return additional_fields


class SshdService(BaseService):
    """Class for parsing sshd messages in the auth log."""

    RE_SSHD_PORTREGEX = re.compile(r"port\s(\d+)")
    RE_USER = re.compile(r"for\s([^\s]+)")

    @classmethod
    def parse(cls, message: str) -> dict[str, str | int]:
        """Parse message from sshd"""
        additional_fields = {}
        if ip_address := RE_IPV4_ADDRESS.search(message):
            field_name = "host_ip" if "listening" in message else "remote_ip"
            additional_fields[field_name] = ip_address.group(0)
        if port := cls.RE_SSHD_PORTREGEX.search(message):
            additional_fields["port"] = int(port.group(1))
        if user := cls.RE_USER.search(message):
            additional_fields["user"] = user.group(1)
        # Accepted publickey for test_user from 8.8.8.8 IP port 12345 ssh2: RSA SHA256:123456789asdfghjklertzuio
        if "Accepted publickey" in message:
            ssh_protocol, encryption_algo, key_info = message.split()[-3:]
            hash_algo, key_hash = key_info.split(":")
            additional_fields["ssh_protocol"] = ssh_protocol.strip(":")
            additional_fields["encryption_algorithm"] = encryption_algo
            additional_fields["hash_algorithm"] = hash_algo
            additional_fields["key_hash"] = key_hash
        if (failed := "Failed" in message) or "Accepted" in message:
            action_type = "failed" if failed else "accepted"
            additional_fields["action"] = f"{action_type} authentication"
            additional_fields["authentication_type"] = "password" if "password" in message else "publickey"

        return additional_fields


class SystemdLogindService(BaseService):
    """Class for parsing systemd-logind messages in the auth log."""

    RE_SYSTEMD_LOGIND_WATCHING = re.compile(
        r"""
        (?P<action>Watching\ssystem\sbuttons)\s # Action is "Watching system buttons"
        on\s(?P<device>[^\s]+)\s                # The device the button is related to -> /dev/input/event0
        \((?P<device_name>.*?)\)                # The device (button) name -> (Power button)
        """,
        re.VERBOSE,
    )

    @classmethod
    def parse(cls, message: str):
        """Parse auth log message from systemd-logind."""
        additional_fields = {}
        # Example: Nov 14 07:14:09 ubuntu-1 systemd-logind[4]: Removed session 4.
        if "Removed" in message:
            additional_fields["action"] = "removed session"
            additional_fields["session"] = message.split()[-1].strip(".")
        elif "Watching" in message and (match := cls.RE_SYSTEMD_LOGIND_WATCHING.search(message)):
            for key, value in match.groupdict().items():
                additional_fields[key] = value
        # Example: New session 4 of user sampleuser.
        elif "New session" in message:
            parts = message.removeprefix("New session ").split()
            additional_fields["action"] = "new session"
            additional_fields["session"] = parts[0]
            additional_fields["user"] = parts[-1].strip(".")
        # Example: Session 4 logged out. Waiting for processes to exit.
        elif "logged out" in message:
            session = message.removeprefix("Session ").split(maxsplit=1)[0]
            additional_fields["action"] = "logged out session"
            additional_fields["session"] = session
        # Example: New seat seat0.
        elif "New seat" in message:
            seat = message.split()[-1].strip(".")
            additional_fields["action"] = "new seat"
            additional_fields["seat"] = seat

        return additional_fields


class SuService(BaseService):
    """Class for parsing su messages in the auth log."""

    RE_SU_BY = re.compile(r"by\s([^\s]+)")
    RE_SU_ON = re.compile(r"on\s([^\s]+)")
    RE_SU_COMMAND = re.compile(r"'(.*?)'")

    @classmethod
    def parse(cls, message: str) -> dict[str, str]:
        additional_fields = {}
        if user := RE_USER.search(message):
            additional_fields["user"] = user.group(1)
        if by := cls.RE_SU_BY.search(message):
            additional_fields["by"] = by.group(1)
        if on := cls.RE_SU_ON.search(message):
            additional_fields["device"] = on.group(1)
        if command := cls.RE_SU_COMMAND.search(message):
            additional_fields["command"] = command.group(1)
        if (failed := "failed" in message) or "Successful" in message:
            additional_fields["su_result"] = "failed" if failed else "success"

        return additional_fields


class PkexecService(BaseService):
    """Class for parsing pkexec messages in the auth log."""

    RE_PKEXEC_COMMAND = re.compile(
        r"""
        (?P<user>\S+?):\sExecuting\scommand\s   # Starts with actual user -> user:
        \[USER=(?P<effective_user>[^\]]+)\]\s   # The impersonated user -> [USER=root]
        \[TTY=(?P<tty>[^\]]+)\]\s               # The tty -> [TTY=unknown]
        \[CWD=(?P<cwd>[^\]]+)\]\s               # Current working directory -> [CWD=/home/user]
        \[COMMAND=(?P<command>[^\]]+)\]         # Command -> [COMMAND=/usr/lib/example]
        """,
        re.VERBOSE,
    )

    @classmethod
    def parse(cls, message: str) -> dict[str, str]:
        """Parse auth log message from pkexec"""
        additional_fields = {}
        if exec_cmd := cls.RE_PKEXEC_COMMAND.search(message):
            additional_fields["action"] = "executing command"
            for key, value in exec_cmd.groupdict().items():
                if value and value.isdigit():
                    value = int(value)
                additional_fields[key] = value

        return additional_fields


class PamUnixService(BaseService):
    RE_PAM_UNIX = re.compile(
        r"""
        pam_unix\([^\s]+:session\):\s(?P<action>session\s\w+)\s     # Session action, usually opened or closed
        for\suser\s(?P<user>[^\s\(]+)(?:\(uid=(?P<user_uid>\d+)\))? # User may contain uid like: root(uid=0)
        (?:\sby\s\(uid=(?P<by_uid>\d+)\))?$                         # Opened action also contains by
        """,
        re.VERBOSE,
    )

    @classmethod
    def parse(cls, message):
        """Parse auth log message from pluggable authentication modules (PAM)."""
        if not (match := cls.RE_PAM_UNIX.search(message)):
            return {}

        additional_fields = {}
        for key, value in match.groupdict().items():
            if value and value.isdigit():
                value = int(value)
            additional_fields[key] = value

        return additional_fields


class AuthLogRecordBuilder:
    """Class for dynamically creating auth log records."""

    RECORD_NAME = "linux/log/auth"
    SERVICES: dict[str, BaseService] = {
        "su": SuService,
        "sudo": SudoService,
        "sshd": SshdService,
        "systemd-logind": SystemdLogindService,
        "pkexec": PkexecService,
    }

    def __init__(self, target: Target):
        self._create_event_descriptor = lru_cache(4096)(self._create_event_descriptor)
        self.target = target

    def _parse_additional_fields(self, service: str | None, message: str) -> dict[str, Any]:
        """Parse additional fields in the message based on the service."""
        if "pam_unix(" in message:
            return PamUnixService.parse(message)

        if service not in self.SERVICES:
            self.target.log.debug("Service %s is not recognised, no additional fields could be parsed", service)
            return {}

        try:
            return self.SERVICES[service].parse(message)
        except Exception as e:
            self.target.log.warning("Parsing additional fields in message '%s' for service %s failed", message, service)
            self.target.log.debug("", exc_info=e)
            raise e

    def build_record(self, ts: datetime, source: Path, line: str) -> TargetRecordDescriptor:
        """Builds an ``AuthLog`` event record."""

        record_fields = [
            ("datetime", "ts"),
            ("path", "source"),
            ("string", "service"),
            ("varint", "pid"),
            ("string", "message"),
        ]

        record_values = {
            "ts": ts,
            "message": line,
            "service": None,
            "pid": None,
            "source": source,
            "_target": self.target,
        }

        match = RE_LINE.search(line)
        if match:
            record_values.update(match.groupdict())

        for key, value in self._parse_additional_fields(record_values["service"], line).items():
            record_type = "string"
            if isinstance(value, int):
                record_type = "varint"

            record_fields.append((record_type, key))
            record_values[key] = value

        # tuple conversion here is needed for lru_cache
        desc = self._create_event_descriptor(tuple(record_fields))
        return desc(**record_values)

    def _create_event_descriptor(self, record_fields) -> TargetRecordDescriptor:
        return TargetRecordDescriptor(self.RECORD_NAME, record_fields)


class AuthPlugin(Plugin):
    """Unix authentication log plugin."""

    def __init__(self, target: Target):
        super().__init__(target)
        self._auth_log_builder = AuthLogRecordBuilder(target)

    def check_compatible(self) -> None:
        var_log = self.target.fs.path("/var/log")
        if not any(var_log.glob("auth.log*")) and not any(var_log.glob("secure*")):
            raise UnsupportedPluginError("No auth log files found")

    @alias("securelog")
    @export(record=DynamicDescriptor(["datetime", "path", "string"]))
    def authlog(self) -> Iterator[Any]:
        """Yield contents of ``/var/log/auth.log*`` and ``/var/log/secure*`` files.

        Order of returned events is not guaranteed to be chronological because of year
        rollover detection efforts for log files without a year in the timestamp.

        The following timestamp formats are recognised automatically. This plugin
        assumes that no custom ``date_format`` template is set in ``syslog-ng`` or ``systemd``
        configuration (defaults to ``M d H:M:S``).

        ISO formatted authlog entries are parsed as can be found in Ubuntu 24.04 and later.

        .. code-block:: text

            CentOS format: Jan 12 13:37:00 hostname daemon: message
            Debian format: Jan 12 13:37:00 hostname daemon[pid]: pam_unix(daemon:session): message
            Ubuntu  24.04: 2024-01-12T13:37:00.000000+02:00 hostname daemon[pid]: pam_unix(daemon:session): message

        Resources:
            - https://help.ubuntu.com/community/LinuxLogFiles
        """

        tzinfo = self.target.datetime.tzinfo

        var_log = self.target.fs.path("/var/log")
        for auth_file in chain(var_log.glob("auth.log*"), var_log.glob("secure*")):
            if is_iso_fmt(auth_file):
                iterable = iso_readlines(auth_file)
            else:
                iterable = year_rollover_helper(auth_file, RE_TS, "%b %d %H:%M:%S", tzinfo)

            for ts, line in iterable:
                yield self._auth_log_builder.build_record(ts, auth_file, line)


def iso_readlines(file: Path) -> Iterator[tuple[datetime, str]]:
    """Iterator reading the provided auth log file in ISO format. Mimics ``year_rollover_helper`` behaviour."""
    with open_decompress(file, "rt") as fh:
        for line in fh:
            if not (match := RE_TS_ISO.match(line)):
                log.warning("No timestamp found in one of the lines in %s!", file)
                log.debug("Skipping line: %s", line)
                continue

            try:
                ts = datetime.strptime(match[0], "%Y-%m-%dT%H:%M:%S.%f%z")
            except ValueError as e:
                log.warning("Unable to parse ISO timestamp in line: %s", line)
                log.debug("", exc_info=e)
                continue

            yield ts, line


def is_iso_fmt(file: Path) -> bool:
    """Determine if the provided auth log file uses new ISO format logging or not."""
    return any(itertools.islice(iso_readlines(file), 0, 2))



import re
from datetime import datetime, timezone
from typing import Iterator

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.fsutil import TargetPath
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export
from dissect.target.target import Target

MssqlErrorlogRecord = TargetRecordDescriptor(
    "microsoft/sql/errorlog",
    [
        ("datetime", "ts"),
        ("string", "instance"),
        ("string", "process"),
        ("string", "message"),
        ("path", "path"),
    ],
)

RE_TIMESTAMP_PATTERN = re.compile(r"^\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}.\d{2}")


class MssqlPlugin(Plugin):
    """Return information related to Microsoft SQL Server.

    Currently returns ERRORLOG messages. These log files contain information such as:
        - Logon failures
        - Enabling/disabling of features, such as xp_cmdshell

    References:
        - https://learn.microsoft.com/en-us/sql/relational-databases/logs/view-offline-log-files
    """

    __namespace__ = "mssql"

    MSSQL_KEY = "HKLM\\SOFTWARE\\Microsoft\\Microsoft SQL Server"
    FILE_GLOB = "ERRORLOG*"

    def __init__(self, target: Target):
        super().__init__(target)
        self.instances = self._find_instances()

    def check_compatible(self) -> None:
        if not self.instances:
            raise UnsupportedPluginError("System does not seem to be running SQL Server")

    @export(record=MssqlErrorlogRecord)
    def errorlog(self) -> Iterator[MssqlErrorlogRecord]:
        """Return all Microsoft SQL Server ERRORLOG messages.

        These log files contain information such as:
         - Logon failures
         - Enabling/disabling of features, such as xp_cmdshell

        Yields MssqlErrorlogRecord instances with fields:

        .. code-block:: text

            ts (datetime): Timestamp of the log line.
            instance (str): SQL Server instance name.
            process (str): Process name.
            message (str): Log message.
            path (Path): Path to the log file.

        References:
            - https://learn.microsoft.com/en-us/sql/relational-databases/logs/view-offline-log-files
        """

        for instance, log_path in self.instances:
            for errorlog in log_path.glob(self.FILE_GLOB):
                # The errorlog includes a BOM, so endianess gets determined automatically
                fh = errorlog.open(mode="rt", encoding="utf-16", errors="surrogateescape")
                buf = ""

                for line in fh:
                    if ts := RE_TIMESTAMP_PATTERN.match(line):
                        yield MssqlErrorlogRecord(
                            ts=datetime.strptime(ts.group(), "%Y-%m-%d %H:%M:%S.%f").replace(tzinfo=timezone.utc),
                            instance=instance,
                            # The process name is a fixed-width field and is always 12 characters long.
                            process=buf[23:35].strip(),
                            message=buf[35:].strip(),
                            path=errorlog,
                            _target=self.target,
                        )
                        buf = ""

                    buf += line

    def _find_instances(self) -> list[str, TargetPath]:
        instances = []

        for subkey in self.target.registry.key(self.MSSQL_KEY).subkeys():
            if subkey.name.startswith("MSSQL") and "." in subkey.name:
                instances.append(
                    (
                        subkey.name,
                        self.target.fs.path(subkey.subkey("SQLServerAgent").value("ErrorLogFile").value).parent,
                    )
                )
        return instances



from dissect.target.helpers.localeutil import normalize_language, normalize_timezone


def test_helpers_localeutil_normalize_timezone():
    assert normalize_timezone("W. Europe Standard Time") == "Europe/Berlin"
    assert normalize_timezone("GMT Standard Time") == "Europe/London"
    assert normalize_timezone("UTC") == "Etc/UTC"


def test_helpers_localeutil_normalize_language():
    assert normalize_language("en_US.UTF-8") == "en_US"
    assert normalize_language("nl") == "nl_NL"



from __future__ import annotations

import logging
import re
import urllib.parse
from datetime import datetime, timezone, tzinfo
from enum import Enum
from pathlib import Path
from typing import BinaryIO, Callable, Iterator, Optional, TypeVar, Union

from dissect.util.ts import from_unix

from dissect.target.helpers import fsutil

log = logging.getLogger(__name__)


def findall(buf: bytes, needle: bytes) -> Iterator[int]:
    offset = 0
    while True:
        offset = buf.find(needle, offset)
        if offset == -1:
            break

        yield offset
        offset += 1


T = TypeVar("T")


def to_list(value: T | list[T]) -> list[T]:
    """Convert a single value or a list of values to a list.

    Args:
        value: The value to convert.

    Returns:
        A list of values.
    """
    if not isinstance(value, list):
        return [value]
    return value


class StrEnum(str, Enum):
    """Sortable and serializible string-based enum"""


def list_to_frozen_set(function: Callable) -> Callable:
    def wrapper(*args):
        args = [frozenset(x) if isinstance(x, list) else x for x in args]
        return function(*args)

    return wrapper


def parse_path_uri(path: Path) -> tuple[Optional[str], Optional[str], Optional[str]]:
    if path is None:
        return None, None, None
    parsed_path = urllib.parse.urlparse(str(path))
    parsed_query = urllib.parse.parse_qs(parsed_path.query, keep_blank_values=True)
    return parsed_path.scheme, parsed_path.path, parsed_query


def parse_options_string(options: str) -> dict[str, Union[str, bool]]:
    result = {}
    for opt in options.split(","):
        if "=" in opt:
            key, _, value = opt.partition("=")
            result[key] = value
        else:
            result[opt] = True
    return result


SLUG_RE = re.compile(r"[/\\ ]")


def slugify(name: str) -> str:
    """Return name with all slashes '/', backslashes '\\' and spaces ' ' replaced by underscores '_'.

    This is useful to turn a name into something that can be used as filename.
    """
    return SLUG_RE.sub("_", name)


def readinto(buffer: bytearray, fh: BinaryIO) -> int:
    """A readinto implementation that uses ``read()``.

    Reads the length of the buffer from ``fh``, and fills the buffer with said data.

    Args:
        buffer: The buffer we read the data into.
        fh: The file-like object we use for reading.

    Returns:
        the size in bytes that was read.
    """
    data = fh.read(len(buffer))
    size = len(data)
    buffer[:size] = data
    return size


STRIP_RE = re.compile(r"^[\s\x00]*|[\s\x00]*$")


def year_rollover_helper(
    path: Path, re_ts: Union[str, re.Pattern], ts_format: str, tzinfo: tzinfo = timezone.utc
) -> Iterator[tuple[datetime, str]]:
    """Helper function for determining the correct timestamps for log files without year notation.

    Supports compressed files by using :func:`open_decompress`.

    Args:
        path: A path to the log file to parse.
        re_ts: Regex pattern for extracting the timestamp from each line.
        ts_format: Time format specification for parsing the timestamp.
        tzinfo: The timezone to use when parsing timestamps.

    Returns:
        An iterator of tuples of the parsed timestamp and the lines of the file in reverse.
    """
    # Convert the mtime to the local timezone so that we get the correct year
    mtime = path.stat().st_mtime
    current_year = from_unix(mtime).astimezone(tzinfo).year
    last_seen_month = None

    with fsutil.open_decompress(path, "rt") as fh:
        warned = False
        for line in fsutil.reverse_readlines(fh):
            line = STRIP_RE.sub(r"", line)
            if not line:
                continue

            timestamp = re.search(re_ts, line)

            if not timestamp:
                if not warned:
                    log.warning("No timestamp found in one of the lines in %s!", path)
                    warned = True
                log.debug("Skipping line: %s", line)
                continue

            try:
                relative_ts = datetime.strptime(timestamp.group(0), ts_format)
            except ValueError as e:
                log.warning("Timestamp '%s' does not match format '%s', skipping line.", timestamp.group(0), ts_format)
                log.debug("", exc_info=e)
                continue

            if last_seen_month and relative_ts.month > last_seen_month:
                current_year -= 1
            last_seen_month = relative_ts.month

            yield relative_ts.replace(year=current_year, tzinfo=tzinfo), line



import argparse
from unittest.mock import MagicMock, call, patch

import pytest

from dissect.target import Target
from dissect.target.report import (
    ExecutionReport,
    TargetExecutionReport,
    format_target_report,
    make_cli_args_overview,
    make_plugin_import_errors_overview,
)
from dissect.target.target import Event


@pytest.fixture
def test_target():
    return Target("test_target")


@pytest.fixture
def func_execs():
    return {"exec2", "exec1"}


@pytest.fixture
def target_execution_report(test_target, func_execs):
    return TargetExecutionReport(
        target=test_target,
        func_execs=func_execs,
    )


@pytest.fixture
def incompatible_plugins():
    return {"incomp_plugin2", "incomp_plugin1"}


@pytest.fixture
def add_incompatible_plugins(target_execution_report, incompatible_plugins):
    for plugin in incompatible_plugins:
        target_execution_report.add_incompatible_plugin(plugin)


@pytest.fixture
def registered_plugins():
    return {"regist_plugin2", "regist_plugin1"}


@pytest.fixture
def add_registered_plugins(target_execution_report, registered_plugins):
    for plugin in registered_plugins:
        target_execution_report.add_registered_plugin(plugin)


@pytest.fixture
def func_errors():
    return {
        "func1": "trace1",
        "func2": "trace2",
    }


@pytest.fixture
def add_func_errors(target_execution_report, func_errors):
    for func, trace in func_errors.items():
        target_execution_report.add_func_error(func, trace)


def test_target_execution_report_add_incompatible_plugin(
    target_execution_report,
    add_incompatible_plugins,
    incompatible_plugins,
):
    for plugin in target_execution_report.incompatible_plugins:
        assert plugin in incompatible_plugins


def test_target_execution_report_add_registered_plugin(
    target_execution_report,
    add_registered_plugins,
    registered_plugins,
):
    for plugin in target_execution_report.registered_plugins:
        assert plugin in registered_plugins


def test_target_execution_report_add_func_error(
    target_execution_report,
    add_func_errors,
    func_errors,
):
    for func, trace in target_execution_report.func_errors.items():
        assert func_errors.get(func) == trace


def test_target_execution_report_as_dict(
    test_target,
    target_execution_report,
    add_incompatible_plugins,
    add_registered_plugins,
    add_func_errors,
    incompatible_plugins,
    registered_plugins,
    func_errors,
    func_execs,
):
    report_dict = target_execution_report.as_dict()
    assert report_dict.get("target") == str(test_target)
    assert report_dict.get("incompatible_plugins") == sorted(incompatible_plugins)
    assert report_dict.get("registered_plugins") == sorted(registered_plugins)
    assert report_dict.get("func_errors") == func_errors
    assert report_dict.get("func_execs") == sorted(func_execs)


@pytest.fixture
def execution_report():
    return ExecutionReport()


@pytest.fixture
def cli_args():
    return argparse.Namespace(foo="bar", baz="bla")


@pytest.fixture
def set_cli_args(execution_report, cli_args):
    execution_report.set_cli_args(cli_args)


@pytest.fixture
def plugin_stats():
    return {
        "_failed": [
            {
                "module": "plugin1",
                "stacktrace": "trace1",
            },
            {
                "module": "plugin2",
                "stacktrace": "trace2",
            },
        ]
    }


@pytest.fixture
def set_plugin_stats(execution_report, plugin_stats):
    execution_report.set_plugin_stats(plugin_stats)


@pytest.fixture
def target1():
    return Target("test1")


@pytest.fixture
def target2():
    return Target("test2")


@pytest.fixture
def target_report1(execution_report, target1):
    return execution_report.add_target_report(target1)


@pytest.fixture
def target_report2(execution_report, target2):
    return execution_report.add_target_report(target2)


@pytest.fixture
def plugin1():
    plugin1 = MagicMock()
    plugin1.__module__ = "test_module"
    plugin1.__qualname__ = "plugin1"
    return plugin1


def test_execution_report_set_cli_args(
    execution_report,
    set_cli_args,
    cli_args,
):
    assert execution_report.cli_args == vars(cli_args)


def test_execution_report_set_plugin_stats(
    execution_report,
    set_plugin_stats,
    plugin_stats,
):
    failed_plugins = plugin_stats["_failed"]
    assert len(execution_report.plugin_import_errors) == len(failed_plugins)

    for failed_plugin in failed_plugins:
        module = failed_plugin["module"]
        stacktrace = failed_plugin["stacktrace"]
        assert execution_report.plugin_import_errors.get(module) == stacktrace


def test_execution_report_get_formatted_report(
    execution_report,
    target_report1,
    target_report2,
):
    with patch("dissect.target.report.make_cli_args_overview", return_value="line_1"):
        with patch("dissect.target.report.make_plugin_import_errors_overview", return_value="line_2"):
            with patch("dissect.target.report.format_target_report", return_value="line_x"):
                assert execution_report.get_formatted_report() == "line_1\nline_2\nline_x\nline_x"


def test_execution_report_add_target_report(
    execution_report,
    target_report1,
    target_report2,
):
    assert len(execution_report.target_reports) == 2
    assert target_report1 in execution_report.target_reports
    assert target_report2 in execution_report.target_reports


def test_execution_report_get_target_report(
    execution_report,
    target_report1,
    target_report2,
    target1,
    target2,
):
    assert target_report1 == execution_report.get_target_report(target1)
    assert target_report2 == execution_report.get_target_report(target2)
    target3 = Target("nope")
    assert execution_report.get_target_report(target3) is None
    target_report3 = execution_report.get_target_report(target3, create=True)
    assert target_report3.target == target3


def test_execution_report__get_plugin_name(execution_report, plugin1):
    assert execution_report._get_plugin_name(plugin1) == "test_module.plugin1"


def test_execution_report_log_incompatible_plugin_plugin_cls(
    execution_report,
    target_report1,
    target1,
    plugin1,
):
    execution_report.log_incompatible_plugin(target1, None, plugin_cls=plugin1)
    assert "test_module.plugin1" in target_report1.incompatible_plugins


def test_execution_report_log_incompatible_plugin_plugin_desc(
    execution_report,
    target_report1,
    target1,
):
    plugin_desc = {"fullname": "test_module.plugin1"}
    execution_report.log_incompatible_plugin(target1, None, plugin_desc=plugin_desc)
    assert "test_module.plugin1" in target_report1.incompatible_plugins


def test_execution_report_log_registered_plugin(
    execution_report,
    target_report1,
    target1,
):
    execution_report.log_registered_plugin(target1, None, plugin_inst=MagicMock())
    assert "unittest.mock.MagicMock" in target_report1.registered_plugins


def test_execution_report_log_func_error(
    execution_report,
    target_report1,
    target1,
    func_errors,
):
    func, trace = next(iter(func_errors.items()))
    execution_report.log_func_error(target1, None, func, trace)
    assert target_report1.func_errors.get(func) == trace


def test_execution_report_log_func_execution(
    execution_report,
    target_report1,
    target1,
    func_execs,
):
    func = next(iter(func_execs))
    execution_report.log_func_execution(target1, None, func)
    assert func in target_report1.func_execs


def test_execution_report_set_event_callbacks(execution_report):
    mock_target = MagicMock()
    event_callbacks = (
        (Event.INCOMPATIBLE_PLUGIN, execution_report.log_incompatible_plugin),
        (Event.REGISTERED_PLUGIN, execution_report.log_registered_plugin),
        (Event.FUNC_EXEC, execution_report.log_func_execution),
        (Event.FUNC_EXEC_ERROR, execution_report.log_func_error),
    )

    execution_report.set_event_callbacks(mock_target)

    assert len(mock_target.mock_calls) == len(event_callbacks)
    for event_type, event_callback in event_callbacks:
        mock_call = call.set_event_callback(event_type=event_type, event_callback=event_callback)
        assert mock_call in mock_target.mock_calls


def test_execution_report_as_dict(
    execution_report,
    set_plugin_stats,
    plugin_stats,
    target_report1,
    target_report2,
    set_cli_args,
    cli_args,
):
    expected_dict = {
        "plugin_import_errors": {
            "plugin1": "trace1",
            "plugin2": "trace2",
        },
        "target_reports": ["report1", "report2"],
        "cli_args": {
            "foo": "bar",
            "baz": "bla",
        },
    }

    with patch.object(target_report1, "as_dict", return_value="report1"):
        with patch.object(target_report2, "as_dict", return_value="report2"):
            execution_report_dict = execution_report.as_dict()
            assert execution_report_dict == expected_dict


def test_report_make_cli_args_overview(
    execution_report,
    set_cli_args,
    cli_args,
):
    cli_args_overview = make_cli_args_overview(execution_report)
    assert "foo: bar" in cli_args_overview
    assert "baz: bla" in cli_args_overview


def test_report_make_plugin_import_errors_overview(
    execution_report,
    set_plugin_stats,
    plugin_stats,
):
    plugin_import_errors_overview = make_plugin_import_errors_overview(execution_report)
    assert "plugin1:\n        trace1" in plugin_import_errors_overview
    assert "plugin2:\n        trace2" in plugin_import_errors_overview


def test_report_format_target_report(
    test_target,
    target_execution_report,
    add_incompatible_plugins,
    add_registered_plugins,
    add_func_errors,
    incompatible_plugins,
    registered_plugins,
    func_errors,
    func_execs,
):
    target_report = format_target_report(target_execution_report)
    assert str(test_target) in target_report

    for registered_plugin in registered_plugins:
        assert registered_plugin in target_report

    for incompatible_plugin in incompatible_plugins:
        assert incompatible_plugin in target_report

    for func, trace in func_errors.items():
        assert func in target_report
        assert trace in target_report

    for func_exec in func_execs:
        assert func_exec in target_report



import hashlib
from pathlib import Path

import pytest

from dissect.target import Target
from dissect.target.helpers import keychain
from dissect.target.loaders.ab import AndroidBackupLoader
from dissect.target.loaders.ab import main as ab_main
from tests._utils import absolute_path


def test_ab_loader(target_bare: Target) -> None:
    ab_path = Path(absolute_path("_data/loaders/ab/test.ab"))
    keychain.register_wildcard_value("password")

    assert AndroidBackupLoader.detect(ab_path)

    loader = AndroidBackupLoader(ab_path)
    loader.map(target_bare)

    assert len(target_bare.filesystems) == 1

    assert list(map(str, target_bare.fs.path("/").rglob("*"))) == [
        "/data",
        "/data/data",
        "/data/data/org.fedorahosted.freeotp",
        "/data/data/org.fedorahosted.freeotp/shared_preferences",
        "/data/data/org.fedorahosted.freeotp/shared_preferences/tokenBackup.xml",
    ]

    buf = target_bare.fs.path("/data/data/org.fedorahosted.freeotp/shared_preferences/tokenBackup.xml").read_bytes()
    assert hashlib.sha1(buf).hexdigest() == "7177d340414d5ca2a835603c347e64e3d2625e47"


def test_ab_unwrapper(tmp_path: Path, monkeypatch: pytest.MonkeyPatch, capsys: pytest.CaptureFixture) -> None:
    ab_path = Path(absolute_path("_data/loaders/ab/test.ab"))

    with monkeypatch.context() as m:
        m.setattr("sys.argv", ["python", str(ab_path), "-p", "password", "-o", str(tmp_path)])
        ab_main()

        assert [entry.name for entry in tmp_path.iterdir()] == ["test.plain.ab"]

        m.setattr("sys.argv", ["python", str(ab_path), "-p", "password", "-o", str(tmp_path), "-t"])
        ab_main()

        assert sorted([entry.name for entry in tmp_path.iterdir()]) == ["test.plain.ab", "test.tar"]



import re
from datetime import datetime
from pathlib import Path
from typing import Iterator

from dissect.target import plugin
from dissect.target.exceptions import FileNotFoundError, UnsupportedPluginError
from dissect.target.helpers.fsutil import open_decompress
from dissect.target.plugins.apps.webserver.webserver import (
    WebserverAccessLogRecord,
    WebserverPlugin,
)
from dissect.target.target import Target

LOG_REGEX = re.compile(
    r'(?P<remote_ip>.*?) - (?P<remote_user>.*?) \[(?P<datetime>\d{2}\/[A-Za-z]{3}\/\d{4}:\d{2}:\d{2}:\d{2} (\+|\-)\d{4})\] "(?P<method>.*?) (?P<uri>.*?) ?(?P<protocol>HTTP\/.*?)?" (?P<status_code>\d{3}) (?P<bytes_sent>-|\d+) (["](?P<referer>(\-)|(.+))["]) "(?P<useragent>.*?)"',  # noqa: E501
)


class NginxPlugin(WebserverPlugin):
    """Nginx webserver plugin."""

    __namespace__ = "nginx"

    def __init__(self, target: Target):
        super().__init__(target)
        self.log_paths = self.get_log_paths()

    def check_compatible(self) -> None:
        if not len(self.log_paths):
            raise UnsupportedPluginError("No NGINX directories found")

    @plugin.internal
    def get_log_paths(self) -> list[Path]:
        log_paths = []

        # Add any well known default NGINX log locations
        log_paths.extend(self.target.fs.path("/var/log/nginx").glob("access.log*"))

        # Check for custom paths in NGINX install config
        if (config_file := self.target.fs.path("/etc/nginx/nginx.conf")).exists():
            for line in config_file.open("rt"):
                line = line.strip()
                if not line or "access_log " not in line:
                    continue

                try:
                    line = line.split("access_log")[1].strip()
                    log_path = self.target.fs.path(line.split()[0])
                    log_paths.extend(
                        path for path in log_path.parent.glob(f"{log_path.name}*") if path not in log_paths
                    )
                except IndexError:
                    self.target.log.warning("Unexpected NGINX log configuration: %s (%s)", line, config_file)

        return log_paths

    @plugin.export(record=WebserverAccessLogRecord)
    def access(self) -> Iterator[WebserverAccessLogRecord]:
        """Return contents of NGINX access log files in unified WebserverAccessLogRecord format.

        References:
            - https://docs.nginx.com/nginx/admin-guide/monitoring/logging/#access_log
            - http://nginx.org/en/docs/http/ngx_http_log_module.html#log_format
        """
        for path in self.log_paths:
            try:
                path = path.resolve(strict=True)
                for line in open_decompress(path, "rt"):
                    line = line.strip()
                    if not line:
                        continue

                    match = LOG_REGEX.match(line)
                    if not match:
                        self.target.log.warning("Could not match NGINX regex format for log line: %s (%s)", line, path)
                        continue

                    log = match.groupdict()
                    yield WebserverAccessLogRecord(
                        ts=datetime.strptime(log["datetime"], "%d/%b/%Y:%H:%M:%S %z"),
                        remote_ip=log["remote_ip"],
                        remote_user=log["remote_user"],
                        method=log["method"],
                        uri=log["uri"],
                        protocol=log["protocol"],
                        status_code=log["status_code"],
                        bytes_sent=log["bytes_sent"].strip("-") or 0,
                        referer=log["referer"],
                        useragent=log["useragent"],
                        source=path,
                        _target=self.target,
                    )
            except FileNotFoundError:
                self.target.log.warning("NGINX log file configured but could not be found (dead symlink?): %s", path)
            except Exception as e:
                self.target.log.warning("An error occured parsing NGINX log file %s: %s", path, str(e))
                self.target.log.debug("", exc_info=e)



from __future__ import annotations

import io
import json
import logging
import re
import sys
from collections import deque
from configparser import ConfigParser, MissingSectionHeaderError
from dataclasses import dataclass
from fnmatch import fnmatch
from types import TracebackType
from typing import (
    Any,
    Callable,
    ItemsView,
    Iterable,
    Iterator,
    KeysView,
    Literal,
    Optional,
    TextIO,
    Union,
)

from defusedxml import ElementTree

from dissect.target.exceptions import ConfigurationParsingError, FileNotFoundError
from dissect.target.filesystem import FilesystemEntry
from dissect.target.helpers.fsutil import TargetPath

try:
    from ruamel.yaml import YAML

    HAS_YAML = True
except ImportError:
    HAS_YAML = False

try:
    if sys.version_info < (3, 11):
        import tomli as toml
    else:
        # tomllib is included since python 3.11
        import tomllib as toml  # novermin

    HAS_TOML = True
except ImportError:
    HAS_TOML = False


log = logging.getLogger(__name__)


def _update_dictionary(current: dict[str, Any], key: str, value: Any) -> None:
    if prev_value := current.get(key):
        if isinstance(prev_value, dict):
            # We can assume the value would be a dict too here.
            prev_value.update(value)
            return

        if isinstance(prev_value, str):
            prev_value = [prev_value]

        if isinstance(prev_value, list):
            # We want to append ``value`` to prev_value
            prev_value.append(value)

    current[key] = prev_value or value


class PeekableIterator:
    # https://more-itertools.readthedocs.io/en/stable/_modules/more_itertools/more.html#peekable

    def __init__(self, iterable):
        self._iterator = iter(iterable)
        self._cache = deque()

    def __iter__(self):
        return self

    def __next__(self):
        if self._cache:
            return self._cache.popleft()

        return next(self._iterator)

    def peek(self):
        if not self._cache:
            try:
                self._cache.append(next(self._iterator))
            except StopIteration:
                return
        return self._cache[0]


class ConfigurationParser:
    """A configuration parser where you can configure certain aspects of the parsing mechanism.

    Args:
        collapse: A ``bool`` or an ``Iterator``:
          If ``True``: it will collapse all the resulting dictionary values.
          If an ``Iterable`` it will collapse on the keys defined in ``collapse``.
        collapse_inverse: Inverses the collapsing mechanism. Collapse on everything that is not inside ``collapse``.
        separator: Contains what values it should look for as a separator.
        comment_prefixes: Contains what constitutes as a comment.
    """

    def __init__(
        self,
        collapse: Union[bool, Iterable[str]] = False,
        collapse_inverse: bool = False,
        separator: tuple[str] = ("=",),
        comment_prefixes: tuple[str] = (";", "#"),
    ) -> None:
        self.collapse_all = collapse is True
        self.collapse = set(collapse) if isinstance(collapse, Iterable) else set()
        self._collapse_check = self._key_not_in_collapse if collapse_inverse else self._key_in_collapse

        self.separator = separator
        self.comment_prefixes = comment_prefixes
        self.parsed_data = {}

    def __getitem__(self, item: Any) -> Union[dict, str]:
        return self.parsed_data[item]

    def __contains__(self, item: str) -> bool:
        return item in self.parsed_data

    def _collapse_dict(self, dictionary: dict, collapse: bool = False) -> dict[str, dict]:
        new_dictionary = {}

        if isinstance(dictionary, list) and collapse:
            return dictionary[-1]

        if not hasattr(dictionary, "items"):
            return dictionary

        for key, value in dictionary.items():
            value = self._collapse_dict(value, self.collapse_all or self._collapse_check(key))
            new_dictionary.update({key: value})

        return new_dictionary

    def _key_in_collapse(self, key: str) -> bool:
        return key in self.collapse

    def _key_not_in_collapse(self, key: str) -> bool:
        return key not in self.collapse

    def parse_file(self, fh: TextIO) -> None:
        """Parse the contents of ``fh`` into key/value pairs.

        This function should **set** :attr:`parsed_data` as a side_effect.

        Args:
            fh: The text to parse.
        """
        raise NotImplementedError()

    def get(self, item: str, default: Optional[Any] = None) -> Any:
        return self.parsed_data.get(item, default)

    def read_file(self, fh: TextIO | io.BytesIO) -> None:
        """Parse a configuration file.

        Raises:
            ConfigurationParsingError: If any exception occurs during during the parsing process.
        """

        try:
            self.parse_file(fh)
        except Exception as e:
            raise ConfigurationParsingError(e.args) from e

        if self.collapse_all or self.collapse:
            self.parsed_data = self._collapse_dict(self.parsed_data)

        if not isinstance(self.parsed_data, dict):
            self.parsed_data = self._collapse_dict(self.parsed_data, False)

    def keys(self) -> KeysView:
        return self.parsed_data.keys()

    def items(self) -> ItemsView:
        return self.parsed_data.items()


class Default(ConfigurationParser):
    """Parse a configuration file specified by ``separator`` and ``comment_prefixes``.

    This parser splits only on the first ``separator`` it finds:

    .. code-block::

        key<separator>value     -> {"key": "value"}

        key<separator>value\n
          continuation
                                -> {"key": "value continuation"}

        # Unless we collapse values, we add them to a list to not overwrite any values.
        key<separator>value1
        key<separator>value2
                                -> {key: [value1, value2]}

        <empty_space><comment>  -> skip
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.SEPARATOR = re.compile(rf"\s*[{''.join(self.separator)}]\s*")
        self.COMMENTS = re.compile(rf"\s*[{''.join(self.comment_prefixes)}]")
        self.skip_lines = self.comment_prefixes + ("\n",)

    def line_reader(self, fh: TextIO, strip_comments: bool = True) -> Iterator[str]:
        for line in fh:
            if line.strip().startswith(self.skip_lines) or not line.strip():
                continue

            if strip_comments:
                line, *_ = self.COMMENTS.split(line, 1)

            yield line

    def parse_file(self, fh: TextIO) -> None:
        information_dict = {}

        prev_key = None
        for line in self.line_reader(fh):
            if line.startswith((" ", "\t")):
                # This part was indented so it is a continuation of the previous key
                prev_value = information_dict.get(prev_key)
                information_dict[prev_key] = " ".join([prev_value, line.strip()])
                continue

            prev_key, *value = self.SEPARATOR.split(line, 1)
            value = value[0].strip() if value else ""

            _update_dictionary(information_dict, prev_key, value)

        self.parsed_data = information_dict


class CSVish(Default):
    """Parses CSV-ish config files (does not confirm to CSV standard!)"""

    def __init__(self, *args, fields: tuple[str], **kwargs) -> None:
        self.fields = fields
        self.num_fields = len(self.fields)
        self.maxsplit = self.num_fields - 1
        super().__init__(*args, **kwargs)

    def parse_file(self, fh: TextIO) -> None:
        information_dict = {}

        for i, raw_line in enumerate(self.line_reader(fh, strip_comments=True)):
            line = raw_line.strip()
            columns = re.split(self.SEPARATOR, line, maxsplit=self.maxsplit)

            if len(columns) < self.num_fields:
                # keep unparsed lines separate (often env vars)
                data = {"line": line}
            else:
                data = dict(zip(self.fields, columns))

            information_dict[str(i)] = data

        self.parsed_data = information_dict


class Ini(ConfigurationParser):
    """Parses an ini file according using the built-in python ConfigParser"""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.parsed_data = ConfigParser(
            strict=False,
            delimiters=self.separator,
            comment_prefixes=self.comment_prefixes,
            allow_no_value=True,
            interpolation=None,
        )
        self.parsed_data.optionxform = str

    def parse_file(self, fh: io.TextIO) -> None:
        offset = fh.tell()
        try:
            self.parsed_data.read_file(fh)
            return
        except MissingSectionHeaderError:
            pass

        fh.seek(offset)
        open_file = io.StringIO("[DEFAULT]\n" + fh.read())
        self.parsed_data.read_file(open_file)


class Txt(ConfigurationParser):
    """Read the file into ``content``, and show the bumber of bytes read."""

    def parse_file(self, fh: TextIO) -> None:
        # Cast the size to a string, to print it out later.
        self.parsed_data = {"content": fh.read(), "size": str(fh.tell())}


class Bin(ConfigurationParser):

    """Read the file into ``binary`` and show the number of bytes read"""

    def parse_file(self, fh: io.BytesIO) -> None:
        self.parsed_data = {"binary": fh.read(), "size": str(fh.tell())}


class Xml(ConfigurationParser):
    """Parses an XML file. Ignores any constructor parameters passed from ``ConfigurationParser``."""

    def _tree(self, tree: ElementTree, root: bool = False) -> dict:
        """Very simple but robust xml -> dict implementation, see comments."""
        nodes = {}
        result = {}
        counter = {}

        # each node is a folder (so the structure is always the same! [1])
        for node in tree.findall("*"):
            # if a node contains multiple nodes with the same name, number them
            if node.tag in counter:
                counter[node.tag] += 1
                nodes[f"{node.tag}-{counter[node.tag]}"] = self._tree(node)
            else:
                counter[node.tag] = 1
                nodes[node.tag] = self._tree(node)

        # all attribs go in the attribute folder
        # (i.e. stable, does not change depending on xml structure! [2]
        # Also, this way we "know" they have been attributes, i.e. we don't lose information! [3]
        if tree.attrib:
            result["attributes"] = tree.attrib

        # all subnodes go in the nodes folder
        if nodes:
            result["nodes"] = nodes

        # content goes into the text folder
        # we don't use special prefixes ($) because XML docs may use them anyway (even though they are forbidden)
        if tree.text:
            if text := tree.text.strip(" \n\r"):
                result["text"] = text

        # if you need to store meta-data, you can extend add more entries here... CDATA, Comments, errors
        result = {tree.tag: result} if root else result
        return result

    def _fix(self, content: str, position: tuple(int, int)) -> str:
        """Quick heuristic fix. If there is an invalid token, just remove it."""
        lineno, offset = position
        lines = content.split("\n")

        line = lines[lineno - 1]
        line = line[: offset - 1] + "" + line[offset + 1 :]

        lines[lineno - 1] = line

        return "\n".join(lines)

    def parse_file(self, fh: TextIO) -> None:
        content = fh.read()
        document = content
        errors = 0
        limit = 20
        tree = {}

        while not tree and errors < limit:
            try:
                tree = self._tree(ElementTree.fromstring(document), root=True)
                break
            except ElementTree.ParseError as err:
                errors += 1
                document = self._fix(document, err.position)

        if not tree:
            # Error limit reached. Thus we consider the document not parseable.
            raise ConfigurationParsingError(f"Could not parse XML file: {fh.name} after {errors} attempts.")

        self.parsed_data = tree


class ListUnwrapper:
    """Provides utility functions to unwrap dictionary objects out of lists."""

    @staticmethod
    def unwrap(data: Union[dict, list]) -> Union[dict, list]:
        """Transforms a list with dictionaries to a dictionary.

        The order of the list is preserved. If no dictionary is found, the list remains untouched:

        .. code-block::

            ["value1", "value2"]    -> ["value1", "value2"]

            {"data": "value"}       -> {"data": "value"}

            [{"data": "value"}]     -> {
                                           "list_item0": {
                                                "data": "value"
                                           }
                                       }
        """
        orig = ListUnwrapper._unwrap_dict_list(data)
        return ListUnwrapper._unwrap_dict(orig)

    @staticmethod
    def _unwrap_dict(data: Union[dict, list]) -> Union[dict, list]:
        """Looks for dictionaries and unwraps its values."""

        if not isinstance(data, dict):
            return data

        root = dict()
        for key, value in data.items():
            _value = ListUnwrapper._unwrap_dict_list(value)
            if isinstance(_value, dict):
                _value = ListUnwrapper._unwrap_dict(_value)
            root[key] = _value

        return root

    @staticmethod
    def _unwrap_dict_list(data: Union[dict, list]) -> Union[dict, list]:
        """Unwraps a list containing dictionaries."""
        if not isinstance(data, list) or not any(isinstance(obj, dict) for obj in data):
            return data

        return_value = {}
        for idx, elem in enumerate(data):
            return_value[f"list_item{idx}"] = elem

        return return_value


class Json(ConfigurationParser):
    """Parses a JSON file."""

    def parse_file(self, fh: TextIO):
        parsed_data = json.load(fh)
        self.parsed_data = ListUnwrapper.unwrap(parsed_data)


class Yaml(ConfigurationParser):
    """Parses a Yaml file."""

    def parse_file(self, fh: TextIO) -> None:
        if HAS_YAML:
            parsed_data = YAML(typ="safe").load(fh)
            self.parsed_data = ListUnwrapper.unwrap(parsed_data)
        else:
            raise ConfigurationParsingError("Failed to parse file, please install ruamel.yaml.")


class Toml(ConfigurationParser):
    """Parses a Toml file."""

    def parse_file(self, fh: TextIO) -> None:
        if HAS_TOML:
            self.parsed_data = toml.loads(fh.read())
        else:
            raise ConfigurationParsingError("Failed to parse file, please install tomli.")


class Env(ConfigurationParser):
    """Parses ``.env`` file contents according to Docker and bash specification.

    Does not apply interpolation of substituted values, eg. ``foo=${bar}`` and does not attempt
    to parse list or dict strings. Does not support dynamic env files, eg. `` foo=`bar` ``. Also
    does not support multi-line key/value assignments (yet).

    Resources:
        - https://docs.docker.com/compose/environment-variables/variable-interpolation/#env-file-syntax
        - https://github.com/theskumar/python-dotenv/blob/main/src/dotenv/parser.py
    """

    RE_KV = re.compile(r"^(?P<key>.+?)=(?P<value>(\".+?\")|(\'.+?\')|(.*?))?(?P<comment> \#.+?)?$")

    def __init__(self, comments: bool = True, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.comments = comments
        self.parsed_data: dict | tuple[dict, str | None] = {}

    def parse_file(self, fh: TextIO) -> None:
        for line in fh.readlines():
            # Blank lines are ignored.
            # Lines beginning with ``#`` are processed as comments and ignored.
            if not line or line[0] == "#" or "=" not in line:
                continue

            # Each line represents a key-value pair. Values can optionally be quoted.
            # Inline comments for unquoted values must be preceded with a space.
            # Value may be empty.
            match = self.RE_KV.match(line)

            # Line could be invalid
            if not match:
                log.warning("Could not parse line in %s: '%s'", fh, line)
                continue

            key = match.groupdict()["key"]
            value = match.groupdict().get("value") or ""
            value = value.strip()
            comment = match.groupdict().get("comment")
            comment = comment.replace(" # ", "", 1) if comment else None

            # Surrounding whitespace characters are removed, unless quoted.
            if value and ((value[0] == '"' and value[-1] == '"') or (value[0] == "'" and value[-1] == "'")):
                is_quoted = True
                value = value.strip("\"'")
            else:
                is_quoted = False
                value = value.strip()

            # Unquoted values may start with a quote if they are properly escaped.
            if not is_quoted and value[:2] in ["\\'", '\\"']:
                value = value[1:]

            # Interpret boolean values
            if value.lower() in ["1", "true"]:
                value = True
            elif value.lower() in ["0", "false"]:
                value = False

            # Interpret integer values
            if isinstance(value, str) and re.match(r"^[0-9]{1,}$", value):
                value = int(value)

            if key.strip() in self.parsed_data:
                log.warning("Duplicate environment key '%s' in file %s", key.strip(), fh)

            self.parsed_data[key.strip()] = (value, comment) if self.comments else value


class ScopeManager:
    """A (context)manager for dictionary scoping.

    This class provides utility functions to keep track of scopes inside a dictionary.

    Attributes:
        _parents: A dictionary accounting what child belongs to which parent dictionary.
        _root: The initial dictionary.
        _current: The current dictionary.
        _previous: The node before the current (changed) node.
    """

    def __init__(self) -> None:
        self._parents = {}
        self._root = {}
        self._current = self._root
        self._previous = None

    def __enter__(self) -> ScopeManager:
        return self

    def __exit__(
        self,
        type: Optional[type[BaseException]],
        value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        self.clean()

    def _set_prev(self, keep_prev: bool) -> None:
        """Set :attr:`_previous` before :attr:`_current` changes."""
        if not keep_prev:
            self._previous = self._current

    def push(self, name: str, keep_prev: bool = False) -> Literal[True]:
        """Push a new key to the :attr:`_current` dictionary and return that we did."""
        child = self._current.get(name, {})

        parent = self._current
        self._parents[id(child)] = parent
        parent[name] = child
        self._set_prev(keep_prev)
        self._current = child
        return True

    def pop(self, keep_prev: bool = False) -> bool:
        """Pop :attr:`_current` and return whether we changed the :attr:`_parents` dictionary."""
        if new_current := self._parents.pop(id(self._current), None):
            self._set_prev(keep_prev)
            self._current = new_current
            return True
        return False

    def update(self, key: str, value: str) -> None:
        """Update the :attr:`_current` dictionary with ``key`` and ``value``."""
        _update_dictionary(self._current, key, value)

    def update_prev(self, key: str, value: str) -> None:
        """Update the :attr:`_previous` dictionary with ``key`` and ``value``."""
        _update_dictionary(self._previous, key, value)

    def is_root(self) -> bool:
        """Utility function to check whether the current dictionary is a root dictionary."""
        return id(self._current) == id(self._root)

    def clean(self) -> None:
        """Clean up the internal state.
        This is called automatically when :class:`ScopeManager` is used as a contextmanager.
        """
        self._parents = {}
        self._root = {}
        self._current = self._root
        self._previous = None


class Indentation(Default):
    """This parser is used for files that use a single level of indentation to specify a different scope.

    Examples of these files are the ``sshd_config`` file.
    Where "Match" statements use a single layer of indentation to specify a scope for the key value pairs.

    The parser parses this as the following:

    .. code-block::

      key value
        key2 value2
                       -> {"key value": {"key2": "value2"}}
    """

    def _parse_line(self, line: str) -> tuple[str, str]:
        key, *value = self.SEPARATOR.split(line.strip(), 1)
        value = value[0].strip() if value else ""
        return key, value

    def _change_scope(
        self,
        manager: ScopeManager,
        line: str,
        key: str,
        next_line: Optional[str] = None,
    ) -> bool:
        """A function to check whether to create a new scope, or go back to a previous one.

        Args:
            manager: A :class:`ScopeManager` that contains the logic to ``push`` and ``pop`` scopes. And keeps state.
            line: The line to be parsed.
            key: The key that should be updated during a :method:`ScopeManager.push`.
            next_line: The next line to be parsed.

        Returns:
            Whether the scope changed or not.
        """
        empty_space = (" ", "\t")
        changed = False

        if next_line is None:
            return False

        if not line.startswith(empty_space):
            changed = manager.pop()

        if not line.startswith(empty_space) and next_line.startswith(empty_space):
            return manager.push(key)
        return changed

    def parse_file(self, fh: TextIO) -> None:
        iterator = PeekableIterator(self.line_reader(fh))
        prev_key = None

        with ScopeManager() as manager:
            for line in iterator:
                key, value = self._parse_line(line)
                changed = self._change_scope(
                    manager=manager,
                    line=line,
                    key=line.strip(),
                    next_line=iterator.peek(),
                )

                if changed:
                    prev_key = line.strip()
                    continue

                if not value:
                    key, value = prev_key, key
                    manager.pop()

                manager.update(key, value)

            self.parsed_data = manager._root


class SystemD(Indentation):
    """A :class:`ConfigurationParser` that specifically parses systemd configuration files.

    Examples:

        .. code-block::

            >>> systemd_data = textwrap.dedent(
                    '''
                    [Section1]
                    Key=Value
                    [Section2]
                    Key2=Value 2\\
                        Value 2 continued
                    '''
                )
            >>> parser = SystemD(io.StringIO(systemd_data))
            >>> parser.parser_items
            {
                "Section1": {
                    "Key": "Value
                },
                "Section2": {
                    "Key2": "Value2 Value 2 continued
                }
            }
    """

    def _change_scope(
        self,
        manager: ScopeManager,
        line: str,
        key: str,
        next_line: Optional[str] = None,
    ) -> bool:
        scope_char = ("[", "]")
        changed = False
        if line.startswith(scope_char):
            if not manager.is_root():
                changed = manager.pop()
            stripped_characters = "".join(scope_char)
            changed = manager.push(key.strip(stripped_characters), changed)

        return changed

    def parse_file(self, fh: TextIO) -> None:
        prev_values = []
        prev_key = None

        with ScopeManager() as manager:
            for line in self.line_reader(fh, strip_comments=False):
                changed = self._change_scope(
                    manager=manager,
                    line=line,
                    key=line.strip(),
                )

                if changed:
                    # Current part is a section header.
                    if prev_values:
                        # Update previous key/value... someone configured it wrong
                        prev_values, prev_key = self._update_continued_values(
                            func=manager.update_prev,
                            key=prev_key,
                            values=prev_values,
                        )
                    continue

                key, value = self._parse_line(line)

                continued_value = value or key
                if continued_value.endswith("\\"):
                    prev_key = prev_key or key
                    prev_values.append(continued_value.strip("\\ "))
                    continue

                if prev_values:
                    prev_values, prev_key = self._update_continued_values(
                        func=manager.update,
                        key=prev_key,
                        values=prev_values + [continued_value],
                    )
                    continue

                manager.update(key, value)

            self.parsed_data = manager._root

    def _update_continued_values(self, func: Callable, key, values: list[str]) -> tuple[list, None]:
        value = " ".join(values)
        func(key, value)
        return [], None


@dataclass(frozen=True)
class ParserOptions:
    collapse: Optional[Union[bool, set]] = None
    collapse_inverse: Optional[bool] = None
    separator: Optional[tuple[str]] = None
    comment_prefixes: Optional[tuple[str]] = None


@dataclass(frozen=True)
class ParserConfig:
    parser: type[ConfigurationParser] = Default
    collapse: Optional[Union[bool, set]] = None
    collapse_inverse: Optional[bool] = None
    separator: Optional[tuple[str]] = None
    comment_prefixes: Optional[tuple[str]] = None
    fields: Optional[tuple[str]] = None

    def create_parser(self, options: Optional[ParserOptions] = None) -> ConfigurationParser:
        kwargs = {}

        for field_name in ["collapse", "collapse_inverse", "separator", "comment_prefixes", "fields"]:
            value = getattr(options, field_name, None) or getattr(self, field_name)
            if value:
                kwargs.update({field_name: value})

        return self.parser(**kwargs)


MATCH_MAP: dict[str, ParserConfig] = {
    "*/systemd/*": ParserConfig(SystemD),
    "*/sysconfig/network-scripts/ifcfg-*": ParserConfig(Default),
    "*/sysctl.d/*.conf": ParserConfig(Default),
    "*/xml/*": ParserConfig(Xml),
    "*.bashrc": ParserConfig(Txt),
    "*/vim/vimrc*": ParserConfig(Txt),
}

CONFIG_MAP: dict[tuple[str, ...], ParserConfig] = {
    "ini": ParserConfig(Ini),
    "xml": ParserConfig(Xml),
    "json": ParserConfig(Json),
    "yml": ParserConfig(Yaml),
    "yaml": ParserConfig(Yaml),
    "cnf": ParserConfig(Default),
    "conf": ParserConfig(Default, separator=(r"\s",)),
    "sample": ParserConfig(Txt),
    "sh": ParserConfig(Txt),
    "key": ParserConfig(Txt),
    "crt": ParserConfig(Txt),
    "pem": ParserConfig(Txt),
    "pl": ParserConfig(Txt),  # various admin panels
    "lua": ParserConfig(Txt),  # wireshark etc.
    "txt": ParserConfig(Txt),
    "systemd": ParserConfig(SystemD),
    "template": ParserConfig(Txt),
    "toml": ParserConfig(Toml),
}


KNOWN_FILES: dict[str, type[ConfigurationParser]] = {
    "ulogd.conf": ParserConfig(Ini),
    "sshd_config": ParserConfig(Indentation, separator=(r"\s",)),
    "hosts.allow": ParserConfig(Default, separator=(":",), comment_prefixes=("#",)),
    "hosts.deny": ParserConfig(Default, separator=(":",), comment_prefixes=("#",)),
    "hosts": ParserConfig(Default, separator=(r"\s",)),
    "nsswitch.conf": ParserConfig(Default, separator=(":",)),
    "lsb-release": ParserConfig(Default),
    "catalog": ParserConfig(Xml),
    "ld.so.cache": ParserConfig(Bin),
    "fstab": ParserConfig(
        CSVish,
        separator=(r"\s",),
        comment_prefixes=("#",),
        fields=("device", "mount", "type", "options", "dump", "pass"),
    ),
    "crontab": ParserConfig(
        CSVish,
        separator=(r"\s",),
        comment_prefixes=("#",),
        fields=("minute", "hour", "day", "month", "weekday", "user", "command"),
    ),
    "shadow": ParserConfig(
        CSVish,
        separator=(r"\:",),
        comment_prefixes=("#",),
        fields=(
            "username",
            "password",
            "lastchange",
            "minpassage",
            "maxpassage",
            "warning",
            "inactive",
            "expire",
            "rest",
        ),
    ),
    "passwd": ParserConfig(
        CSVish,
        separator=(r"\:",),
        comment_prefixes=("#",),
        fields=("username", "password", "uid", "gid", "gecos", "homedir", "shell"),
    ),
    "mime.types": ParserConfig(CSVish, separator=(r"\s+",), comment_prefixes=("#",), fields=("name", "extensions")),
}


def parse(path: Union[FilesystemEntry, TargetPath], hint: Optional[str] = None, *args, **kwargs) -> ConfigurationParser:
    """Parses the content of an ``path`` or ``entry`` to a dictionary.

    Args:
        path: The path to either a directory or file.
        hint: What kind of parser should be used.
        collapse: Whether it should collapse everything or just a certain set of keys.
        collapse_inverse: Invert the collapse function to collapse everything but the keys inside ``collapse``.
        separator: The separator that should be used for parsing.
        comment_prefixes: What is specified as a comment.

    Raises:
        FileNotFoundError: If the ``path`` is not a file.
    """

    entry = path
    if isinstance(path, TargetPath):
        entry = path.get()

    if not entry.is_file(follow_symlinks=True):
        raise FileNotFoundError(f"Could not parse {path} as a dictionary.")

    options = ParserOptions(*args, **kwargs)

    return parse_config(entry, hint, options)


def parse_config(
    entry: FilesystemEntry,
    hint: Optional[str] = None,
    options: Optional[ParserOptions] = None,
) -> ConfigurationParser:
    parser_type = _select_parser(entry, hint)

    parser = parser_type.create_parser(options)
    with entry.open() as fh:
        if not isinstance(parser, Bin):
            open_file = io.TextIOWrapper(fh, encoding="utf-8")
        else:
            open_file = io.BytesIO(fh.read())
        parser.read_file(open_file)

    return parser


def _select_parser(entry: FilesystemEntry, hint: Optional[str] = None) -> ParserConfig:
    if hint and (parser_type := CONFIG_MAP.get(hint)):
        return parser_type

    for match, value in MATCH_MAP.items():
        if fnmatch(entry.path, f"{match}"):
            return value

    extension = entry.path.rsplit(".", 1)[-1]

    extention_parser = CONFIG_MAP.get(extension, ParserConfig(Default))
    return KNOWN_FILES.get(entry.name, extention_parser)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import argparse
import json
import logging
from datetime import datetime
from pathlib import Path
from typing import Union

from dissect.target import Target
from dissect.target.exceptions import TargetError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.tools.query import record_output
from dissect.target.tools.utils import (
    args_to_uri,
    catch_sigpipe,
    configure_generic_arguments,
    process_generic_arguments,
)

InfoRecord = TargetRecordDescriptor(
    "target/info",
    [
        ("datetime", "last_activity"),
        ("datetime", "install_date"),
        ("net.ipaddress[]", "ips"),
        ("string", "os_family"),
        ("string", "os_version"),
        ("string", "architecture"),
        ("string[]", "language"),
        ("string", "timezone"),
        ("string[]", "disks"),
        ("string[]", "volumes"),
        ("string[]", "children"),
    ],
)


log = logging.getLogger(__name__)
logging.lastResort = None
logging.raiseExceptions = False


@catch_sigpipe
def main():
    help_formatter = argparse.ArgumentDefaultsHelpFormatter
    parser = argparse.ArgumentParser(
        description="target-info",
        fromfile_prefix_chars="@",
        formatter_class=help_formatter,
    )
    parser.add_argument("targets", metavar="TARGETS", nargs="*", help="Targets to display info from")
    parser.add_argument("--from-file", nargs="?", type=Path, help="file containing targets to load")
    parser.add_argument("-s", "--strings", action="store_true", help="print output as string")
    parser.add_argument("-r", "--record", action="store_true", help="print output as record")
    parser.add_argument("-j", "--json", action="store_true", help="output records as pretty json")
    parser.add_argument("-J", "--jsonlines", action="store_true", help="output records as one-line json")
    parser.add_argument("-L", "--loader", action="store", default=None, help="select a specific loader (i.e. vmx, raw)")
    configure_generic_arguments(parser)

    args, rest = parser.parse_known_args()

    process_generic_arguments(args)

    if not args.targets and not args.from_file:
        parser.error("too few arguments")

    if args.from_file:
        if not args.from_file.exists():
            parser.error(f"--from-file {args.from_file} does not exist")

        targets = args.from_file.read_text().splitlines()
        while targets[-1] == "":
            targets = targets[:-1]
        args.targets = targets

    targets = args_to_uri(args.targets, args.loader, rest) if args.loader else args.targets

    try:
        for i, target in enumerate(Target.open_all(targets)):
            try:
                if args.jsonlines:
                    print(json.dumps(get_target_info(target), default=str))
                elif args.json:
                    print(json.dumps(get_target_info(target), indent=4, default=str))
                elif args.record:
                    rs = record_output(args.strings)
                    rs.write(InfoRecord(**get_target_info(target), _target=target))
                else:
                    if i > 0:
                        print("-" * 70)
                    print_target_info(target)
            except Exception as e:
                target.log.error("Exception in retrieving information for target: `%s`. Use `-vv` for details.", target)
                target.log.debug("", exc_info=e)
    except TargetError as e:
        log.error(e)
        log.debug("", exc_info=e)
        parser.exit(1)


def get_target_info(target: Target) -> dict[str, Union[str, list[str]]]:
    return {
        "disks": get_disks_info(target),
        "volumes": get_volumes_info(target),
        "children": get_children_info(target),
        "hostname": target.hostname,
        "domain": get_optional_func(target, "domain"),
        "ips": target.ips,
        "os_family": target.os,
        "os_version": target.version,
        "architecture": target.architecture,
        "language": get_optional_func(target, "language"),
        "timezone": get_optional_func(target, "timezone"),
        "install_date": get_optional_func(target, "install_date"),
        "last_activity": get_optional_func(target, "activity"),
    }


def get_optional_func(target: Target, func: str) -> Union[str, None]:
    if target.has_function(func):
        return getattr(target, func)


def print_target_info(target: Target) -> None:
    print(target)

    for name, value in get_target_info(target).items():
        if name in ["disks", "volumes", "children"]:
            if not any(value):
                continue
            print(f"\n{name.capitalize()}")
            for i in value:
                values = " ".join([f'{k}="{v}"' for k, v in i.items()])
                print(f"- <{name.capitalize()[:-1].replace('re', '')} {values}>")
            continue

        if isinstance(value, list):
            value = ", ".join(map(str, value))

        if isinstance(value, datetime):
            value = value.isoformat(timespec="microseconds")

        if name == "hostname":
            print()

        print(f"{name.capitalize().replace('_', ' '):14s} : {value}")


def get_disks_info(target: Target) -> list[dict[str, Union[str, int]]]:
    return [{"type": d.__class__.__name__, "size": d.size} for d in target.disks]


def get_volumes_info(target: Target) -> list[dict[str, Union[str, int]]]:
    return [{"name": v.name, "size": v.size, "fs": v.fs.__class__.__name__} for v in target.volumes]


def get_children_info(target: Target) -> list[dict[str, str]]:
    return [{"type": c.type, "path": str(c.path)} for c in target.list_children()]


if __name__ == "__main__":
    main()
